import logging
import os
import matplotlib.pyplot as plt
from pytplot import get_data
from . import mms_load_mec


def mms_orbit_plot(trange=['2015-10-16', '2015-10-17'],
                   probes=[1, 2, 3, 4],
                   data_rate='srvy',
                   xr=None,
                   yr=None,
                   plane='xy',
                   coord='gse',
                   xsize=5,
                   ysize=5,
                   marker='x',
                   markevery=10,
                   markersize=5,
                   earth=True,
                   dpi=300,
                   save_png='',
                   save_pdf='',
                   save_eps='',
                   save_jpeg='',
                   save_svg='',
                   return_plot_objects=False,
                   display=True
                   ):
    """
    This function creates MMS orbit plots
    
    Parameters:
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probes: list of str
            probe #, e.g., '4' for MMS4

        data_rate: str
            instrument data rate, e.g., 'srvy' or 'brst'

        plane: str
            coordinate plane to plot (options: 'xy', 'yz', 'xz')

        xr: list of float
            two element list specifying x-axis range

        yr: list of float
            two element list specifying y-axis range

        coord: str
            coordinate system

        xsize: float
            size of the figure in the x-direction, in inches (default: 5)

        ysize: float
            size of the figure in the y-direction, in inches (default: 5)

        marker: str
            marker style for the data points (default: 'x')

        markevery: int or sequence of int
            plot a marker at every n-th data point (default: 10)

        markersize: float
            size of the marker in points (default: 5)

        earth: bool
            plot a reference image of the Earth (default: True)

        dpi: int
            dots per inch for the plot (default: 300)

        save_png: str
            file path to save the plot as a PNG file (default: None)

        save_pdf: str
            file path to save the plot as a PDF file (default: None)

        save_eps: str
            file path to save the plot as an EPS file (default: None)

        save_jpeg: str
            file path to save the plot as a JPEG file (default: None)

        save_svg: str
            file path to save the plot as an SVG file (default: None)

        return_plot_objects: bool
            whether to return the plot objects as a tuple (default: False)

        display: bool
            whether to display the plot using matplotlib's `show()` function (default: True)

    """
    spacecraft_colors = [(0,0,0), (213/255,94/255,0), (0,158/255,115/255), (86/255,180/255,233/255)]

    mec_vars = mms_load_mec(trange=trange, data_rate=data_rate, probe=probes, varformat='*_r_' + coord, time_clip=True)

    if len(mec_vars) == 0:
        logging.error('Problem loading MEC data')
        return

    plane = plane.lower()
    coord = coord.lower()

    if plane not in ['xy', 'yz', 'xz']:
        logging.error('Error, invalid plane specified; valid options are: xy, yz, xz')
        return

    if coord not in ['eci', 'gsm', 'geo', 'sm', 'gse', 'gse2000']:
        logging.error('Error, invalid coordinate system specified; valid options are: eci, gsm, geo, sm, gse, gse2000')
        return

    km_in_re = 6371.2

    fig, axis = plt.subplots(sharey=True, sharex=True, figsize=(xsize, ysize))

    if earth:
        im = plt.imread(os.path.dirname(os.path.realpath(__file__)) + '/mec/earth_polar1.png')
        plt.imshow(im, extent=(-1, 1, -1, 1))

    plot_count = 0

    for probe in probes:
        position_data = get_data('mms' + str(probe) + '_mec_r_' + coord)
        if position_data is None:
            logging.error('No ' + data_rate + ' MEC data found for ' + 'MMS' + str(probe))
            continue
        else:
            t, d = position_data
            plot_count += 1

        if plane == 'xy':
            axis.plot(d[:, 0]/km_in_re, d[:, 1]/km_in_re, label='MMS' + str(probe), color=spacecraft_colors[int(probe)-1], marker=marker, markevery=markevery, markersize=markersize)
            axis.set_xlabel('X Position, Re')
            axis.set_ylabel('Y Position, Re')
        if plane == 'yz':
            axis.plot(d[:, 1]/km_in_re, d[:, 2]/km_in_re, label='MMS' + str(probe), color=spacecraft_colors[int(probe)-1], marker=marker, markevery=markevery, markersize=markersize)
            axis.set_xlabel('Y Position, Re')
            axis.set_ylabel('Z Position, Re')
        if plane == 'xz':
            axis.plot(d[:, 0]/km_in_re, d[:, 2]/km_in_re, label='MMS' + str(probe), color=spacecraft_colors[int(probe)-1], marker=marker, markevery=markevery, markersize=markersize)
            axis.set_xlabel('X Position, Re')
            axis.set_ylabel('Z Position, Re')

        axis.set_aspect('equal')

    if plot_count > 0:  # at least one plot created
        axis.legend()
        axis.set_title(trange[0] + ' to ' + trange[1])
        axis.annotate(coord.upper() + ' coordinates', xy=(0.6, 0.05), xycoords='axes fraction')
        if xr is not None:
            axis.set_xlim(xr)
        if yr is not None:
            axis.set_ylim(yr)

        if return_plot_objects:
            return fig, axis

        if save_png is not None and save_png != '':
            plt.savefig(save_png + '.png', dpi=dpi)

        if save_eps is not None and save_eps != '':
            plt.savefig(save_eps + '.eps', dpi=dpi)

        if save_svg is not None and save_svg != '':
            plt.savefig(save_svg + '.svg', dpi=dpi)

        if save_pdf is not None and save_pdf != '':
            plt.savefig(save_pdf + '.pdf', dpi=dpi)

        if save_jpeg is not None and save_jpeg != '':
            plt.savefig(save_jpeg + '.jpeg', dpi=dpi)

        if display:
            plt.show()



import logging
import numpy as np
from pyspedas import time_string, time_double
from datetime import datetime


def dailynames(directory='', trange=None, res=24*3600., hour_res=False, file_format='%Y%m%d', prefix='', suffix=''):
    '''
    Creates a list of file names using a time range, resoution and file format
    Based on Davin Larson's file_dailynames in IDL SPEDAS

    Parameters:
        directory: str
            String containing the directory for the generated file names

        trange: list of str, list of datetime or list of floats
            Two-element list containing the start and end times for the file names

        res: float
            File name resolution in seconds (default: 24*3600., i.e., daily)

        file_format: str
            Format of the file names using strftime directives; for reference: https://strftime.org
            (default: %Y%m%d, i.e., daily)

        prefix: str
            file name prefix

        suffix: str
            file name suffix

    Returns:
        List containing filenames

    '''
    if trange is None:
        logging.error('No trange specified')
        return

    if hour_res:
        res = 3600.
        file_format = '%Y%m%d%H'

    # allows the user to pass in trange as list of datetime objects
    if type(trange[0]) == datetime and type(trange[1]) == datetime:
        trange = [time_string(trange[0].timestamp()), time_string(trange[1].timestamp())]

    tr = [trange[0], trange[1]]
    
    if isinstance(trange[0], str):
        tr[0] = time_double(trange[0])
    if isinstance(trange[1], str):
        tr[1] = time_double(trange[1])

    # Davin's magic heisted from file_dailynames in IDL
    mmtr = [np.floor(tr[0]/res), np.ceil(tr[1]/res)]

    if mmtr[1]-mmtr[0] < 1:
        n = 1
    else:
        n = int(mmtr[1]-mmtr[0])

    times = [(float(num)+mmtr[0])*res for num in range(n)]

    dates = []
    files = []
        
    for time in times:
        if time_string(time, fmt=file_format) not in dates:
            dates.append(time_string(time, fmt=file_format))
            
    for date in dates:
        files.append(directory + prefix + date + suffix)

    return files



import pyspedas
import unittest

from pyspedas.utilities.data_exists import data_exists
from pyspedas.mms.fpi.mms_fpi_split_tensor import mms_fpi_split_tensor
from pyspedas.mms.fpi.mms_fpi_ang_ang import mms_fpi_ang_ang
from pyspedas.mms.fpi.mms_get_fpi_dist import mms_get_fpi_dist
from pyspedas.mms.fpi.mms_pad_fpi import mms_pad_fpi


class FPITestCases(unittest.TestCase):
    def test_angle_angle(self):
        mms_fpi_ang_ang('2015-10-16/13:06:30', data_rate='brst', save_png='mms1_fpi_ang_ang_brst', display=False)
        mms_fpi_ang_ang('2015-10-16/13:06:30', save_jpeg='mms1_fpi_ang_ang', display=False)
        mms_fpi_ang_ang('2015-10-16/13:06:30', probe='4', save_svg='mms4_fpi_ang_ang', display=False)
        mms_fpi_ang_ang('2015-10-16/13:06:30', probe='4', save_eps='mms4_fpi_ang_ang_viridis', cmap='viridis', display=False)

    def test_pad(self):
        trange = ['2015-10-16/13:06:29', '2015-10-16/13:06:32']
        pyspedas.mms.fpi(trange=trange, data_rate='brst', datatype=['dis-dist', 'des-dist', 'dis-moms'], time_clip=True)
        pyspedas.mms.fgm(trange=trange, data_rate='brst')
        dists = mms_get_fpi_dist('mms1_dis_dist_brst')
        dists_e = mms_get_fpi_dist('mms1_des_dist_brst')
        pa_dist = mms_pad_fpi(dists, trange=trange, mag_data='mms1_fgm_b_gse_brst_l2_bvec')
        pa_dist = mms_pad_fpi(dists_e, trange=trange, mag_data='mms1_fgm_b_gse_brst_l2_bvec')
        pa_dist = mms_pad_fpi(dists, time='2015-10-16/13:06:30', units='eflux', mag_data='mms1_fgm_b_gse_brst_l2_bvec')
        pa_dist = mms_pad_fpi(dists,
                      subtract_bulk=True,
                      time='2015-10-16/13:06:30',
                      units='eflux',
                      mag_data='mms1_fgm_b_gse_brst_l2_bvec',
                      vel_data='mms1_dis_bulkv_gse_brst')

    def test_split_tensors(self):
        data = pyspedas.mms.fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'],
                                data_rate='brst',
                                datatype=['dis-moms', 'des-moms'])

        split_vars = mms_fpi_split_tensor('mms1_dis_temptensor_gse_brst')

        components = ['xx', 'xy', 'xz', 'yx', 'yy', 'yz', 'zx', 'zy', 'zz']
        output = ['mms1_dis_temptensor_gse_brst_' + component for component in components]

        for v in output:
            self.assertTrue(data_exists(v))

        split_vars = mms_fpi_split_tensor('mms1_des_temptensor_gse_brst')

        components = ['xx', 'xy', 'xz', 'yx', 'yy', 'yz', 'zx', 'zy', 'zz']
        output = ['mms1_des_temptensor_gse_brst_' + component for component in components]

        for v in output:
            self.assertTrue(data_exists(v))


if __name__ == '__main__':
    unittest.main()



import os
import unittest
from pyspedas.utilities.data_exists import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_load_mag_data(self):
        out_vars = pyspedas.de2.mag(time_clip=True)
        self.assertTrue(data_exists('bx'))
        self.assertTrue(data_exists('by'))
        self.assertTrue(data_exists('bz'))

    def test_load_nacs_data(self):
        out_vars = pyspedas.de2.nacs()
        self.assertTrue(data_exists('O_density'))
        self.assertTrue(data_exists('N_density'))

    def test_load_rpa_data(self):
        out_vars = pyspedas.de2.rpa()
        self.assertTrue(data_exists('ionDensity'))
        self.assertTrue(data_exists('ionTemperature'))

    def test_load_fpi_data(self):
        out_vars = pyspedas.de2.fpi()
        self.assertTrue(data_exists('TnF'))

    # issue with the CDFs here, 2Nov2022
    # def test_load_idm_data(self):
    #     out_vars = pyspedas.de2.idm()
    #     self.assertTrue(data_exists('ionVelocityY'))
    #     self.assertTrue(data_exists('ionVelocityZ'))

    def test_load_wats_data(self):
        out_vars = pyspedas.de2.wats()
        self.assertTrue(data_exists('density'))
        self.assertTrue(data_exists('Tn'))

    def test_load_vefi_data(self):
        out_vars = pyspedas.de2.vefi()
        self.assertTrue(data_exists('spectA'))
        self.assertTrue(data_exists('spectB'))
        self.assertTrue(data_exists('spectC'))

    def test_load_lang_data(self):
        out_vars = pyspedas.de2.lang()
        self.assertTrue(data_exists('plasmaDensity'))
        self.assertTrue(data_exists('electronTemp'))

    def test_load_notplot(self):
        out_vars = pyspedas.de2.mag(notplot=True)
        self.assertTrue('bz' in out_vars)

    def test_downloadonly(self):
        files = pyspedas.de2.mag(downloadonly=True, trange=['1983-2-16', '1983-2-17'])
        self.assertTrue(os.path.exists(files[0]))


if __name__ == '__main__':
    unittest.main()

    


import os
import unittest
from pyspedas.utilities.data_exists import data_exists
import pyspedas
from pytplot import del_data


class LoadTestCases(unittest.TestCase):
    def tearDown(self):
        del_data('*')

    def test_downloadonly(self):
        files = pyspedas.image.mena(downloadonly=True)
        self.assertTrue(os.path.exists(files[0]))

    def test_load_notplot(self):
        mena_vars = pyspedas.image.mena(notplot=True)
        self.assertTrue('Image0' in mena_vars)

    def test_load_lena_data(self):
        lena_vars = pyspedas.image.lena(time_clip=True)
        self.assertTrue(data_exists('Image0'))

    def test_load_mena_data(self):
        mena_vars = pyspedas.image.mena()
        self.assertTrue(data_exists('Image0'))

    def test_load_hena_data(self):
        hena_vars = pyspedas.image.hena()
        self.assertTrue(data_exists('Image0'))

    def test_load_rpi_data(self):
        rpi_vars = pyspedas.image.rpi()
        self.assertTrue(data_exists('Amplitude'))

    def test_load_euv_data(self):
        euv_vars = pyspedas.image.euv()
        self.assertTrue(data_exists('IMAGE'))

    def test_load_fuv_data(self):
        fuv_vars = pyspedas.image.fuv()
        self.assertTrue(data_exists('WIC_PIXELS'))

    def test_load_orb_data(self):
        orb_vars = pyspedas.image.orbit()
        self.assertTrue(data_exists('GSM_POS'))
        self.assertTrue(data_exists('GSM_VEL'))

    def test_load_preorb_data(self):
        orb_vars = pyspedas.image.orbit(datatype='pre_or')
        self.assertTrue(data_exists('GSM_POS'))
        self.assertTrue(data_exists('GSM_VEL'))


if __name__ == '__main__':
    unittest.main()



import os
import unittest
from pyspedas.utilities.data_exists import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_load_mpa_data(self):
        out_vars = pyspedas.lanl.mpa(time_clip=True)
        self.assertTrue(data_exists('dens_lop'))
        self.assertTrue(data_exists('vel_lop'))
        self.assertTrue(data_exists('temp_lop'))

    def test_load_spa_data(self):
        out_vars = pyspedas.lanl.spa(time_clip=True)
        self.assertTrue(data_exists('spa_p_temp'))
        self.assertTrue(data_exists('spa_e_temp'))
        self.assertTrue(data_exists('spa_a_flux'))
        self.assertTrue(data_exists('spa_e_flx'))
        self.assertTrue(data_exists('spa_p_flx'))

    def test_load_notplot(self):
        out_vars = pyspedas.lanl.spa(notplot=True)
        self.assertTrue('spa_p_temp' in out_vars)

    def test_downloadonly(self):
        files = pyspedas.lanl.spa(downloadonly=True, trange=['2004-10-31', '2004-11-01'])
        self.assertTrue(os.path.exists(files[0]))


if __name__ == '__main__':
    unittest.main()

    


from pyspedas.themis.load import load


def fit(trange=['2007-03-23', '2007-03-24'],
        probe='c',
        level='l2',
        suffix='',
        get_support_data=False,
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads THEMIS FIT data

    Parameters:
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str or list of str
            Spacecraft probe letter(s) ('a', 'b', 'c', 'd' and/or 'e')

        level: str
            Data type; Valid options: 'l1', 'l2'

        suffix: str
            The tplot variable names will be given this suffix.
            By default, no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load
            (if not specified, all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified
            in the trange keyword

    Returns:
        List of tplot variables created.

    """
    return load(instrument='fit', trange=trange, level=level,
                suffix=suffix, get_support_data=get_support_data,
                varformat=varformat, varnames=varnames,
                downloadonly=downloadonly, notplot=notplot,
                probe=probe, time_clip=time_clip, no_update=no_update)


def cal_fit(probe='a', no_cal=False):
    """
    Converts raw FIT parameter data into physical quantities.
    Warning: This function is in debug state

    Currently, it assumes that "th?_fit" variable is already loaded

    Parameters:
        probe: str
            Spacecraft probe letter ('a', 'b', 'c', 'd' and/or 'e')
        no_cal: bool
            If ture do not apply boom shortening factor or Ex offset defaults

    Returns:
        th?_fgs tplot variable
    """
    import math
    import numpy as np
    import logging

    from pytplot import get_data, store_data, tplot_names, options
    from pyspedas.utilities.download import download
    from pyspedas.themis.config import CONFIG
    from pyspedas.utilities.time_double import time_float_one
    from copy import deepcopy
    from numpy.linalg import inv

    if probe == 'f':
        logging.warning(f"Probe f is not supported. Please use IDL version of 'thm_cal_fit' in SPEDAS.")
        return

    # calibration parameters
    lv12 = 49.6  # m
    lv34 = 40.4  # m
    lv56 = 5.6  # m

    # This values provide better agreement with IDL
    lv12 = 49.599997
    lv34 = 40.400003
    lv56 = 5.59999981

    # calibration table
    cpar = {"e12": {"cal_par_time": '2002-01-01/00:00:00',
                    "Ascale": -15000.0 / (lv12 * 2. ** 15.),
                    "Bscale": -15000.0 / (lv12 * 2. ** 15.),
                    "Cscale": -15000.0 / (lv12 * 2. ** 15.),
                    "theta": 0.0,
                    "sigscale": 15000. / (lv12 * 2. ** 15.),
                    "Zscale": -15000. / (lv56 * 2. ** 15.),
                    "units": 'mV/m'},
            "e34": {"cal_par_time": '2002-01-01/00:00:00',
                    "Ascale": -15000.0 / (lv34 * 2. ** 15.),
                    "Bscale": -15000.0 / (lv34 * 2. ** 15.),
                    "Cscale": -15000.0 / (lv34 * 2. ** 15.),
                    "theta": 0.0,
                    "sigscale": 15000. / (lv34 * 2. ** 15.),
                    "Zscale": -15000. / (lv56 * 2. ** 15.),
                    "units": 'mV/m'},
            "b": {"cal_par_time": '2002-01-01/00:00:00',
                  "Ascale": 1.,
                  "Bscale": 1.,
                  "Cscale": 1.,
                  "theta": 0.0,
                  "sigscale": 1.,
                  "Zscale": 1.,
                  "units": 'nT'}}

    # tplot options
    color_str = ['blue', 'green', 'red']
    color_str2 = ['magenta', 'blue', 'cyan', 'green', 'orange']
    b_str = ['Bx', 'By', 'Bz']
    e_str = ['Ex', 'Ey', 'Ez']
    b_units = cpar['b']['units']
    e_units = cpar['e12']['units']
    b_units_str = f'[{b_units}]'
    e_units_str = f'[{e_units}]'
    b_data_att = {'units': b_units, 'cal_par_time': cpar['b']['cal_par_time'], 'data_type': 'calibrated',
                  'coord_sys': 'dsl'}
    b_data_att_sigma = {'units': b_units}
    e_data_att = {'units': e_units, 'cal_par_time': cpar['e12']['cal_par_time'], 'data_type': 'calibrated',
                  'coord_sys': 'dsl'}
    e_data_att_sigma = {'units': e_units}

    b_opt_dict = {'legend_names': b_str, 'ysubtitle': b_units_str, 'color': color_str, 'alpha': 1}
    e_opt_dict = {'legend_names': e_str, 'ysubtitle': e_units_str, 'color': color_str, 'alpha': 1}

    # if tplot does not show 5th legend name, update tplot
    b_opt_dict2 = {'legend_names': ['A', 'B', 'C', 'Sig', '<Bz>'],
                   'ysubtitle': b_units_str, 'color': color_str2, 'alpha': 1}
    e_opt_dict2 = {'legend_names': ['A', 'B', 'C', 'Sig', '<Ez>'],
                   'ysubtitle': e_units_str, 'color': color_str2, 'alpha': 1}

    # Get list of tplot variables
    tnames = tplot_names(True)  # True for quiet output

    # Get data from th?_fit variable
    tvar = 'th' + probe + '_fit'

    # B-field fit (FGM) processing

    if tvar not in tnames:
        logging.warning(f"Variable {tvar} is not found")
        return

    # Using deep copy to create an independent instance
    d = deepcopy(get_data(tvar))  # NOTE: Indexes are not the same as in SPEDAS, e.g. 27888x2x5

    # establish probe number in cal tables
    sclist = {'a': 0, 'b': 1, 'c': 2, 'd': 3, 'e': 4, 'f': -1}  # for probe 'f' no flatsat FGM cal files
    scn = sclist[probe]

    #  Rotation vectors
    rotBxy_angles = [29.95, 29.95, 29.95, 29.95,
                     29.95]  # vassilis 6/2/2007: deg to rotate FIT on spin plane to match DSL on 5/4
    rotBxy = rotBxy_angles[scn]  # vassilis 4/28: probably should be part of CAL table as well...
    cs = math.cos(rotBxy * math.pi / 180)  # vassilis
    sn = math.sin(rotBxy * math.pi / 180)  # vassilis

    adc2nT = 50000. / 2. ** 24  # vassilis 2007 - 04 - 03

    # B - field fit(FGM)
    i = 1

    d.y[:, i, 0] = cpar['b']['Ascale'] * d.y[:, i, 0] * adc2nT  # vassilis
    d.y[:, i, 1] = cpar['b']['Bscale'] * d.y[:, i, 1] * adc2nT  # vassilis
    d.y[:, i, 2] = cpar['b']['Cscale'] * d.y[:, i, 2] * adc2nT  # vassilis
    d.y[:, i, 3] = cpar['b']['sigscale'] * d.y[:, i, 3] * adc2nT  # vassilis
    d.y[:, i, 4] = cpar['b']['Zscale'] * d.y[:, i, 4] * adc2nT  # vassilis

    # Calculating Bzoffset using thx+'/l1/fgm/0000/'+thx+'_fgmcal.txt'
    thx = 'th' + probe
    remote_name = thx + '/l1/fgm/0000/' + thx + '_fgmcal.txt'
    calfile = download(remote_file=remote_name,
                       remote_path=CONFIG['remote_data_dir'],
                       local_path=CONFIG['local_data_dir'],
                       no_download=False)
    if not calfile:
        # This code should never be executed
        logging.warning(f"Calibration file {thx}_fgmcal.txt is not found")
        return

    caldata = np.loadtxt(calfile[0], converters={0: time_float_one})
    # TODO: In SPEDAS we checks if data is already calibrated
    # Limit to the time of interest
    caltime = caldata[:, 0]
    t1 = np.nonzero(caltime < d.times.min())[0]  # left time
    t2 = np.nonzero(caltime <= d.times.max())[0]  # right time

    Bzoffset = np.zeros(d.times.shape)
    if t1.size + t2.size > 1:  # Time range exist in the file
        tidx = np.arange(t1[-1], t2[-1] + 1)
        caltime = caltime[tidx]
        offi = caldata[tidx, 1:4]  # col 1-3
        cali = caldata[tidx, 4:13]  # col 4-13
        # spinperii = caldata[tidx, 13]  # col 14 not in use
        flipxz = -1 * np.fliplr(np.identity(3))
        # SPEDAS: offi2 = invert(transpose([cali[istart, 0:2], cali[istart, 3:5], cali[istart, 6:8]]) ## flipxz)##offi[istart, *]
        for t in range(0, caltime.size):
            offi2 = inv(np.c_[cali[t, 0:3], cali[t, 3:6], cali[t, 6:9]].T @ flipxz) @ offi[t, :]
            tidx = d.times >= caltime[t]
            Bzoffset[tidx] = offi2[2]  # last element

    Bxprime = cs * d.y[:, i, 1] + sn * d.y[:, i, 2]
    Byprime = -sn * d.y[:, i, 1] + cs * d.y[:, i, 2]
    Bzprime = -d.y[:, i, 4] - Bzoffset  # vassilis 4/28 (SUBTRACTING offset from spinaxis POSITIVE direction)

    # d is a namedtuple and does not support direct copy by value
    dprime = deepcopy(d)
    dprime.y[:, i, 1] = Bxprime  # vassilis DSL
    dprime.y[:, i, 2] = Byprime  # vassilis DSL
    dprime.y[:, i, 4] = Bzprime  # vassilis DSL

    # Create fgs variable and remove nans
    fgs = dprime.y[:, i, [1, 2, 4]]
    idx = ~np.isnan(fgs[:, 0])
    fgs_data = {'x': d.times[idx], 'y': fgs[idx, :]}

    # Save fgs tplot variable
    tvar = 'th' + probe + '_fgs'
    store_data(tvar, fgs_data, attr_dict=b_data_att)
    options(tvar, opt_dict=b_opt_dict)

    # Save fgs_sigma variable
    fit_sigma_data = {'x': d.times[idx], 'y': d.y[idx, i, 3]}
    tvar = 'th' + probe + '_fgs_sigma'
    store_data(tvar, fit_sigma_data, attr_dict=b_data_att_sigma)
    options(tvar, opt_dict=b_opt_dict)

    # Save bfit variable
    bfit_data = {'x': d.times[:], 'y': d.y[:, i, :].squeeze()}
    tvar = 'th' + probe + '_fit_bfit'
    store_data(tvar, bfit_data, attr_dict=b_data_att)
    options(tvar, opt_dict=b_opt_dict2)

    # E-field fit (EFI) processing

    # Get data from th?_fit_code variable
    tvar = 'th' + probe + '_fit_code'
    d_code = None  # Blank variable, if d_code is not used

    if tvar in tnames:
        i = 0
        d_code = get_data(tvar)
        e12_ss = (d_code.y[:, i] == int("e1", 16)) | (d_code.y[:, i] == int("e5", 16))
        e34_ss = (d_code.y[:, i] == int("e3", 16)) | (d_code.y[:, i] == int("e7", 16))
    else:
        # Default values (if no code)
        ne12 = d.times.size
        e12_ss = np.ones(ne12, dtype=bool)  # create an index arrays
        e34_ss = np.zeros(ne12, dtype=bool)

    # Save 'efs' datatype before "hard wired" calibrations.
    # An EFI-style calibration is performed below.
    i = 0
    efs = d.y[:, i, [1, 2, 4]]
    # Locate samples with non-NaN data values.  Save the indices in
    # efsx_good, then at the end of calibration, pull the "good"
    # indices out of the calibrated efs[] array to make the thx_efs
    # tplot variable.
    efsx_good = ~np.isnan(efs[:, 0])

    if np.any(efsx_good):
        if np.any(e34_ss):  # rotate efs 90 degrees if necessary, if e34 was used in spinfit
            tmp = d.y[e34_ss, i, :]  # Apply logical arrays
            # TODO: there should be a better way to combine indexes
            efs[e34_ss, :] = tmp[:, [2, 1, 4]]  # Apply dimension selection
            efs[e34_ss, 0] = -efs[e34_ss, 0]

    efsz = d.y[:, i, 4]  # save Ez separately, for possibility that it's the SC potential

    # Use cpar to calibrate
    if np.any(e12_ss):
        d.y[e12_ss, i, 0] = cpar["e12"]["Ascale"] * d.y[e12_ss, i, 0]
        d.y[e12_ss, i, 1] = cpar["e12"]["Bscale"] * d.y[e12_ss, i, 1]
        d.y[e12_ss, i, 2] = cpar["e12"]["Cscale"] * d.y[e12_ss, i, 2]
        d.y[e12_ss, i, 3] = cpar["e12"]["sigscale"] * d.y[e12_ss, i, 3]
        d.y[e12_ss, i, 4] = cpar["e12"]["Zscale"] * d.y[e12_ss, i, 4]
    if np.any(e34_ss):
        d.y[e34_ss, i, 0] = cpar["e34"]["Ascale"] * d.y[e34_ss, i, 0]
        d.y[e34_ss, i, 1] = cpar["e34"]["Bscale"] * d.y[e34_ss, i, 1]
        d.y[e34_ss, i, 2] = cpar["e34"]["Cscale"] * d.y[e34_ss, i, 2]
        d.y[e34_ss, i, 3] = cpar["e34"]["sigscale"] * d.y[e34_ss, i, 3]
        d.y[e34_ss, i, 4] = cpar["e34"]["Zscale"] * d.y[e34_ss, i, 4]

    # save fit_efit variable
    fit_efit_data = {'x': d.times, 'y': d.y[:, i, :]}
    tvar = 'th' + probe + '_fit_efit'
    store_data(tvar, fit_efit_data, attr_dict=e_data_att)
    options(tvar, opt_dict=e_opt_dict2)

    # thx_efs and thx_efs_sigma,
    # Calibrate efs data by applying E12 calibration factors, not despinning, then applying despun (spin-dependent)
    # calibration factors from E12 (the spin-independent offset is subtracted on-board):

    # Load calibration file, e.g. tha/l1/eff/0000/tha_efi_calib_params.txt
    remote_name = thx + '/l1/eff/0000/' + thx + '_efi_calib_params.txt'
    eficalfile = download(remote_file=remote_name,
                          remote_path=CONFIG['remote_data_dir'],
                          local_path=CONFIG['local_data_dir'],
                          no_download=False)

    if not eficalfile:
        # This code should never be executed
        logging.warning(f"Calibration file {thx}_efi_calib_params.txt is not found")
        return

    colnums = {"time": [0], "edc_offset": [14, 15, 16], "edc_gain": [17, 18, 19],
               "BOOM_LENGTH": [26, 27, 28], "BOOM_SHORTING_FACTOR": [29, 30, 31],
               "DSC_OFFSET": [32, 33, 34]}  # List of columns to be loaded
    collist = list()
    [collist.extend(cnum) for cnum in colnums.values()]
    collist.sort()  # ensurer that the list of columns is sorted
    eficaltxt = np.loadtxt(eficalfile[0], skiprows=1, max_rows=1, converters={0: time_float_one}, usecols=collist)
    eficaldata = {"time": eficaltxt[0], "gain": eficaltxt[4:7], "offset": eficaltxt[1:4],
                  "boom_length": eficaltxt[7:10], "boom_shorting_factor": eficaltxt[10:13],
                  "dsc_offset": eficaltxt[13:16]}

    # Boom
    exx = eficaldata["boom_length"]
    if not no_cal:
        exx *= eficaldata["boom_shorting_factor"]

    # Calibrate E field
    # Calibrate Ex and Ey spinfits that are derived from E12 only!
    if np.any(e12_ss):
        efs[e12_ss, 0:2] = -1000. * eficaldata["gain"][0] * efs[e12_ss, 0:2] / exx[0]
    if np.any(e34_ss):
        efs[e34_ss, 0:2] = -1000. * eficaldata["gain"][1] * efs[e34_ss, 0:2] / exx[1]
    # Calibrate Ez spinfit by itself:
    efs[:, 2] = -1000. * eficaldata["gain"][2] * efs[:, 2] / exx[2]

    # DC Offset
    if not no_cal:
        efs -= eficaldata["dsc_offset"]

    # Here, if the fit_code is 'e5'x (229) then efs[*,2] contains the spacecraft potential, so set all of those values
    # to Nan, jmm, 19-Apr-2010
    # Or if the fit_code is 'e7'x (231), this will also be including the SC potential, jmm,22-oct-2010
    if d_code is not None:
        sc_port = (d_code.y[:, i] == int("e5", 16)) | (d_code.y[:, i] == int("e7", 16))
        if np.any(sc_port):
            efs[sc_port, 2] = np.nan

    # save efs variable
    efs_data = {'x': d.times[efsx_good], 'y': efs[efsx_good, :]}  # efs[efsx_good,*]
    tvar = 'th' + probe + '_efs'
    store_data(tvar, efs_data, attr_dict=e_data_att)
    options(tvar, opt_dict=e_opt_dict)

    # save efs_sigma variable
    efs_sigma_data = {'x': d.times[efsx_good], 'y': d.y[efsx_good, i, 3]}  # d.y[efsx_good, 3, idx]
    tvar = 'th' + probe + '_efs_sigma'
    store_data(tvar, efs_sigma_data, attr_dict=e_data_att_sigma)
    options(tvar, opt_dict=e_opt_dict)

    # save efs_0
    efs_0_data = deepcopy(efs)
    efs_0_data[:, 2] = 0
    efs_0 = {'x': d.times[efsx_good], 'y': efs_0_data[efsx_good, :]}
    tvar = 'th' + probe + '_efs_0'
    store_data(tvar, efs_0, attr_dict=e_data_att_sigma)
    options(tvar, opt_dict=e_opt_dict)

    # calculate efs_dot0
    Ez = (efs[:, 0] * fgs[:, 0] + efs[:, 1] * fgs[:, 1]) / (-1 * fgs[:, 2])
    angle = np.arccos(fgs[:, 2] / np.sqrt(np.sum(fgs ** 2, axis=1))) * 180 / np.pi
    angle80 = angle > 80
    if np.any(angle80):
        Ez[angle80] = np.NaN
    efx_dot0_data = deepcopy(efs)
    efx_dot0_data[:, 2] = Ez

    # save efs_dot0
    efs_dot0 = {'x': d.times[efsx_good], 'y': efx_dot0_data[efsx_good, :]}
    tvar = 'th' + probe + '_efs_dot0'
    store_data(tvar, efs_dot0, attr_dict=e_data_att_sigma)
    options(tvar, opt_dict=e_opt_dict)



import os
import datetime
import pandas as pd
import numpy as np
from pyspedas.utilities.download import download
from pyspedas import time_double


def load_leap_table(reload=False):
    """
        Loads the leap second table for converting TAI to unix times

    Parameters
    -----------
        reload: bool
            Re-load the leap second table, even if it exists locally.

            This shouldn't be needed until at least 2035:
                https://www.scientificamerican.com/article/the-leap-seconds-time-is-up-world-votes-to-stop-pausing-clocks/

    Returns
    ---------
        dict containing 'dates' with array of Julian dates corresponding
        to the leap seconds in the 'leaps' array

    """
    if os.environ.get('CDF_LEAPSECONDSTABLE') is not None:
        table_file = os.environ.get('CDF_LEAPSECONDSTABLE')
    elif os.environ.get('SPEDAS_DATA_DIR') is not None:
        table_file = os.path.join(os.environ.get('SPEDAS_DATA_DIR'), 'CDFLeapSeconds.txt')
    else:
        table_file = os.path.join('data', 'CDFLeapSeconds.txt')

    table_dir = os.path.dirname(table_file)

    if reload or not os.path.exists(table_file):
        downloaded = download(remote_path='https://cdf.gsfc.nasa.gov/html/',
                              remote_file='CDFLeapSeconds.txt',
                              local_path=table_dir)

    cols = ['Year', 'Month', 'Day', 'LS', 'Drift']
    table = pd.read_csv(table_file,
                        delim_whitespace=True,
                        dtype=str,
                        names=cols,
                        comment=';',
                        skipinitialspace=True,
                        index_col=False)

    leap_dates = table['Year'].to_numpy() + '-' + table['Month'].to_numpy() + '-' + table['Day'].to_numpy()
    leap_dates = time_double(leap_dates)
    juls = np.array(leap_dates)/86400.0 + datetime.date(1970, 1, 1).toordinal() + 1721424.5

    return {'leaps': np.float64(table['LS'].to_numpy()),
            'juls': juls}



import logging
import pandas as pd
import numpy as np
from pyspedas.analysis.time_clip import time_clip as tclip
from pytplot import store_data


def mms_load_att_tplot(filenames, level='def', probe='1', datatypes=['spinras', 'spindec'], suffix='', trange=None):
    """
    Helper routine for loading state data (ASCII files from the SDC); not meant to be called directly; see pyspedas.mms.state instead
    
    """
    prefix = 'mms' + probe

    file_times = []
    file_lras = []
    file_ldecs = []
    tvalues = {}
    
    logging.info('Loading attitude files can take some time; please be patient...')
    for file in filenames:
        logging.info('Loading ' + file)
        rows = pd.read_csv(file, delim_whitespace=True, header=None, skiprows=49)

        times = rows.shape[0]-1
        time_values = np.empty(times)
        lra_values = np.empty(times)
        ldec_values = np.empty(times)

        for time_idx in range(0, times):
            # these files can overlap, so avoid duplicates
            if tvalues.get(time_values[time_idx]):
                continue
            time_values[time_idx] = pd.to_datetime(rows[0][time_idx], format='%Y-%jT%H:%M:%S.%f').timestamp()
            tvalues[time_values[time_idx]] = 1
            lra_values[time_idx] = rows[13][time_idx]
            ldec_values[time_idx] = rows[14][time_idx]

        file_times.append(time_values)
        file_lras.append(lra_values)
        file_ldecs.append(ldec_values)

    file_times_array = np.concatenate(file_times)
    file_lras_array = np.concatenate(file_lras)
    file_ldecs_array = np.concatenate(file_ldecs)

    file_times_sorted_idx = np.argsort(file_times_array)
    file_times_sorted = file_times_array[file_times_sorted_idx]
    file_lras_sorted = file_lras_array[file_times_sorted_idx]
    file_ldecs_sorted = file_ldecs_array[file_times_sorted_idx]

    file_times_uniq = np.unique(file_times_sorted, return_index=True)
    file_lras_out = file_lras_sorted[file_times_uniq[1]]
    file_ldecs_out = file_ldecs_sorted[file_times_uniq[1]]

    if 'spinras' in datatypes:
        store_data(prefix + '_' + level + 'att_spinras' + suffix, data={'x': file_times_uniq[0], 'y': file_lras_out})
        tclip(prefix + '_' + level + 'att_spinras' + suffix, trange[0], trange[1], suffix='')

    if 'spindec' in datatypes:
        store_data(prefix + '_' + level + 'att_spindec' + suffix, data={'x': file_times_uniq[0], 'y': file_ldecs_out})
        tclip(prefix + '_' + level + 'att_spindec' + suffix, trange[0], trange[1], suffix='')




"""
Find all pytplot names that follow a pattern that may contain a wildcard.

Notes
-----
Allowed wildcards are ? for a single character, * from multiple characters.

"""
import fnmatch
from pytplot import tplot_names


def tnames(pattern=None):
    """
    Find pytplot names.

    Parameters
    ----------
    pattern : str, optional
        Patern to search for. The default is None, which returns all names.

    Returns
    -------
    name_list : list of str
        List of pytplot variables.

    """
    name_list = list()
    all_names = tplot_names(quiet=True)

    if pattern is None:
        name_list.extend(all_names)
    else:
        if isinstance(pattern, str):
            name_list.extend(fnmatch.filter(all_names, pattern))
        else:
            for p in pattern:
                name_list.extend(fnmatch.filter(all_names, p))

    return name_list



import logging
import numpy as np
from pyspedas import tnames, tinterpol
from pytplot import get_data, store_data
from pyspedas.cotrans.quaternions import qtom, mtoq, qslerp
from pyspedas.cotrans.matrix_array_lib import ctv_verify_mats, ctv_left_mats, ctv_swap_hands


def tvector_rotate(mat_var_in, vec_var_in, newname=None):
    """
    Rotates array data by a set of coordinate
    transformation matrices and outputs tplot variables.
    This is designed mainly for use with fac_matrix_make, 
    but can be used for more general purposes.

    Input
    ------
        mat_var_in: str
            Tplot variable containing the transformation matrices

        vec_var_in: str or list of str
            Tplot variables to be transformed

    Parameters
    ----------
        newname: str or list of str
            Name of the output tplot variables

    Returns
    ---------
        Names of the variables created.

    """

    if tnames(mat_var_in) == []:
        logging.error('Transformation requires the matrix variable to be set to a valid tplot variable.')
        return

    if tnames(vec_var_in) == []:
        logging.error('Transformation requires the vector variables to be set to a valid tplot variable.')
        return

    vec_var_in = tnames(vec_var_in)

    if newname is None:
        newname = [vec_var + '_rot' for vec_var in vec_var_in]

    if not isinstance(newname, list):
        newname = [newname]

    if len(newname) != len(vec_var_in):
        logging.error('Length of newname keyword should match the length of vec_var_in')
        return

    out_names = []

    # loop over the vectors
    for vec_var, new_var in zip(vec_var_in, newname):
        vec_data = get_data(vec_var)
        vec_metadata = get_data(vec_var, metadata=True)
        mat_data = get_data(mat_var_in)
        m_d_y = mat_data.y

        if not np.array_equal(vec_data.times, mat_data.times) and len(mat_data.times) != 1:
            verify_check = ctv_verify_mats(mat_data.y)

            is_left_mat = ctv_left_mats(mat_data.y)

            # left-handed matrices can mess up qslerping
            if is_left_mat:
                q_in = mtoq(ctv_swap_hands(mat_data.y))
            else:
                q_in = mtoq(mat_data.y)

            # interpolate quaternions
            q_out = qslerp(q_in, mat_data.times, vec_data.times)

            # turn quaternions back into matrices
            m_d_y = qtom(q_out)

            if is_left_mat:
                m_d_y = ctv_swap_hands(m_d_y)

        vec_fac = np.zeros((len(vec_data.times), len(vec_data.y[0, :])))

        for i in range(0, len(vec_data.times)):
            if m_d_y.shape[0] == 1:  # only a single matrix
                matrix = m_d_y[0, :, :]
            else:
                matrix = m_d_y[i, :, :]

            vec_fac[i, :] = matrix @ vec_data.y[i, :]

        saved = store_data(new_var, data={'x': vec_data.times, 'y': vec_fac}, attr_dict=vec_metadata)

        if saved:
            out_names.append(new_var)

    return out_names



import os
import unittest
from pyspedas.utilities.data_exists import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_downloadonly(self):
        files = pyspedas.stereo.mag(trange=['2013-1-6', '2013-1-7'], downloadonly=True)
        self.assertTrue(os.path.exists(files[0]))

    def test_load_mag_data(self):
        mag_vars = pyspedas.stereo.mag(trange=['2013-11-5', '2013-11-6'], time_clip=True)
        self.assertTrue(data_exists('BFIELD'))
        mag_vars = pyspedas.stereo.mag(trange=['2013-11-5', '2013-11-6'], datatype='32hz')
        mag_vars = pyspedas.stereo.mag(trange=['2013-11-5', '2013-11-6'], notplot=True)
        self.assertTrue('BFIELD' in mag_vars)

    def test_load_swea_data(self):
        swea_vars = pyspedas.stereo.swea(trange=['2013-1-5', '2013-1-6'], time_clip=True)
        self.assertTrue(data_exists('SWEASpectra'))

    def test_load_ste_data(self):
        ste_vars = pyspedas.stereo.ste(trange=['2013-1-5', '2013-1-6'], time_clip=True)
        self.assertTrue(data_exists('STE_spectra'))

    def test_load_sept_data(self):
        sept_vars = pyspedas.stereo.sept(trange=['2013-1-5', '2013-1-6'], time_clip=True)
        self.assertTrue(data_exists('Spec_0_E'))

    def test_load_sit_data(self):
        sit_vars = pyspedas.stereo.sit(trange=['2013-1-5', '2013-1-6'], time_clip=True)
        self.assertTrue(data_exists('H_Intensity'))

    def test_load_let_data(self):
        let_vars = pyspedas.stereo.let(trange=['2013-1-5', '2013-1-6'], time_clip=True)
        self.assertTrue(data_exists('H_Hi_sec_flux'))

    def test_load_het_data(self):
        het_vars = pyspedas.stereo.het(trange=['2013-1-5', '2013-1-6'], time_clip=True)
        self.assertTrue(data_exists('Electron_Flux'))
        self.assertTrue(data_exists('Proton_Flux'))

    def test_load_plastic_data(self):
        p_vars = pyspedas.stereo.plastic(trange=['2013-11-5', '2013-11-6'], probe='b')
        self.assertTrue(data_exists('proton_number_density'))
        self.assertTrue(data_exists('proton_bulk_speed'))
        self.assertTrue(data_exists('proton_temperature'))


if __name__ == '__main__':
    unittest.main()



import os
import logging
import numpy as np
import pandas as pd
from pyspedas import time_string, time_double
from pyspedas.mms.mms_tai2unix import mms_tai2unix
from pyspedas.utilities.download import download
from pyspedas.mms.mms_config import CONFIG


def mms_brst_events(trange=None, reload=False):
    """
    Prints a list of burst mode segment selections from the MMS data segment database

    Parameters
    -----------
        trange: list of str or list of float
            Time range to list burst mode events

        reload:
            Re-download the burst mode events database
    """
    if trange is None:
        logging.error('Time range not specified. ')
        return

    remote_path = 'https://lasp.colorado.edu/mms/sdc/public/service/latis/'
    remote_file = 'mms_burst_data_segment.csv'

    if reload or not os.path.exists(os.path.join(CONFIG['local_data_dir'], remote_file)):
        brst_file = download(remote_path=remote_path, remote_file=remote_file, local_path=CONFIG['local_data_dir'])
        if len(brst_file) > 0:
            brst_file = brst_file[0]
    else:
        brst_file = os.path.join(CONFIG['local_data_dir'], remote_file)

    table = load_csv_file(brst_file)

    descriptions = table['DISCUSSION'].to_numpy()
    authors = table['SOURCEID'].to_numpy()
    start_tai = np.float64(table['TAISTARTTIME'].to_numpy())
    end_tai = np.float64(table['TAIENDTIME'].to_numpy())

    start_unix = mms_tai2unix(start_tai)
    end_unix = mms_tai2unix(end_tai)

    # sort based on start time
    sorted_indices = np.argsort(start_unix)
    descriptions = descriptions[sorted_indices]
    authors = authors[sorted_indices]
    start_unix = start_unix[sorted_indices]
    end_unix = end_unix[sorted_indices]

    trange = time_double(trange)
    idxs = ((start_unix >= trange[0]) & (start_unix <= trange[1])) & ((end_unix <= trange[1]) & (end_unix >= trange[0]))
    indices = np.argwhere(idxs).flatten()
    descriptions = descriptions[indices]
    authors = authors[indices]
    start_times = start_unix[indices]
    end_times = end_unix[indices]

    for desc, author, start_time, end_time in zip(descriptions, authors, start_times, end_times):
        print(time_string(start_time, fmt='%Y-%m-%d/%H:%M:%S') + ' - ' + time_string(end_time, fmt='%Y-%m-%d/%H:%M:%S') + ': ' + str(desc) + ' (' + str(author) + ')')


def load_csv_file(filename, cols=None):
    """
    Loads the burst segment CSV file into a pandas data frame
    """
    if cols is None:
        cols = ['DATASEGMENTID', 'TAISTARTTIME', 'TAIENDTIME', 'PARAMETERSETID', 'FOM', 'ISPENDING', 'INPLAYLIST', 'STATUS', 'NUMEVALCYCLES', 'SOURCEID', 'CREATETIME', 'FINISHTIME', 'OBS1NUMBUFS', 'OBS2NUMBUFS', 'OBS3NUMBUFS', 'OBS4NUMBUFS', 'OBS1ALLOCBUFS', 'OBS2ALLOCBUFS', 'OBS3ALLOCBUFS', 'OBS4ALLOCBUFS', 'OBS1REMFILES', 'OBS2REMFILES', 'OBS3REMFILES', 'OBS4REMFILES', 'DISCUSSION', 'empty1', 'empty2']
    df = pd.read_csv(filename, dtype=str, names=cols, on_bad_lines='skip', skiprows=1)
    return df



import unittest
from pyspedas.utilities.data_exists import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_load_dst_data(self):
        # final
        dst_vars = pyspedas.kyoto.dst(trange=['2015-10-15', '2015-10-16'])
        self.assertTrue(data_exists('kyoto_dst'))
        # provisional
        dst_vars = pyspedas.kyoto.dst(trange=['2019-10-15', '2019-10-16'])
        self.assertTrue(data_exists('kyoto_dst'))
        # real time
        dst_vars = pyspedas.kyoto.dst(trange=['2022-7-15', '2022-7-16'])
        self.assertTrue(data_exists('kyoto_dst'))

    def test_errors(self):
        pyspedas.kyoto.dst(trange=None)
        pyspedas.kyoto.dst(trange=['1015-10-15', '1015-10-16'])


if __name__ == '__main__':
    unittest.main()



"""
Smooths a tplot variable.

Uses a boxcar average of the specified width.

Notes
-----
Similar to tsmooth2.pro in IDL SPEDAS.
Also, see: https://www.harrisgeospatial.com/docs/SMOOTH.html

"""
import logging
import math
import numpy as np
import pyspedas
import pytplot


def smooth(data, width=10, preserve_nans=None):
    """
    Boxcar average.

    Parameters
    ----------
    data : list of floats
        The data should be a one-dim array.
    width : float, optional
        Data window to use for smoothing. The default is 10.
    preserve_nans : bool, optional
        If None, then replace NaNs.

    Returns
    -------
    result : list of floats
        Smoothed data.

    """
    result = data.copy()
    N = len(data)

    if N <= width:
        logging.error("smooth: Not enough points.")
        return result

    for i, d in enumerate(data):
        if (i >= (width-1)/2) and (i <= N-(width+1)/2):
            if (preserve_nans is not None) and data[i] is np.NaN:
                continue
            tsum = 0
            count = 0
            for j in range(int(width)):
                idx = math.ceil(i+j-width/2)
                if data[idx] is not np.NaN:
                    tsum += data[idx]
                    count += 1
            if count > 0:  # otherwise, all NaN
                result[i] = (1/width) * tsum
    return result


def tsmooth(names, width=10, median=None, preserve_nans=None,
            new_names=None, suffix=None, overwrite=None):
    """
    Smooths data.

    Parameters
    ----------
    names: str/list of str
        List of pytplot names.
    width: int, optional
        Data window to use for smoothing. The default is 10.
    median: bool, optional
        Apply the median as well.
    preserve_nans: bool, optional
        If None, then replace NaNs.
    new_names: str/list of str, optional
        List of new_names for pytplot variables.
        If not given, then a suffix is applied.
    suffix: str, optional
        A suffix to apply. Default is '-s'.
    overwrite: bool, optional
        Replace the existing tplot name.

    Returns
    -------
    None.

    """
    old_names = pyspedas.tnames(names)

    if len(old_names) < 1:
        logging.error('tsmooth error: No pytplot names were provided.')
        return

    if suffix is None:
        suffix = '-s'

    if overwrite is not None:
        n_names = old_names
    elif new_names is None:
        n_names = [s + suffix for s in old_names]
    else:
        n_names = new_names

    if isinstance(n_names, str):
        n_names = [n_names]

    if len(n_names) != len(old_names):
        n_names = [s + suffix for s in old_names]

    for i, old in enumerate(old_names):
        new = n_names[i]

        if new != old:
            pyspedas.tcopy(old, new)

        data = pytplot.data_quants[new].values

        dim = data.shape
        if len(dim) == 1:
            data = smooth(data, width=width, preserve_nans=preserve_nans)
        else:
            for k in range(dim[1]):
                data[:, k] = smooth(data[:, k], width=width,
                                    preserve_nans=preserve_nans)

        pytplot.data_quants[new].values = data

        logging.info('tsmooth was applied to: ' + new)



"""
; --- This program calculates the unit vector of sight direction (in SGI
;     coordinates [satellite spin coordinates]) of the  n-th MEPe/MEPi channel.
;
; --- You can find some documents on the geometory of MEPs in the MEP server
;     directory: /raid/meps/docs. The geometory of MEPe explained in
;     MEPe_EPS_r1_black.pdf is used in this function. The document that
;     describes MEPi geometory is missing now (2017/Oct/01).
;
"""

import numpy as np


def get_mepe_az_dir_in_sga(fluxdir=False):
    """
    ;parameter
    ;channel azimuthal angle
    ; --- starts from -Z(SGI) axis and increases toward -Y(SGI) axis.
    ;     Note that MEPe channel number is clockwise when you see from
    ;     +X(SGI) direction.
    """

    azi_8 = -39.4  # ;[deg] offset of channel 8 from -Z(SGI) axis
    d_azi = 22.5  # ;[deg] interval of each center of aperture

    phi_array = azi_8 + (np.arange(16) - 8.) * d_azi

    """
    ;inclination of apertures (elevation)
    ; --- This value is based on visual speculation of Fig. 3 in
    ;     MEPe_EPS_r1_black.pdf.
    """
    theta_array = np.full(shape=(16), fill_value=2.0)

    """
    ;unit vector of sight direction
    ; --- Note that positive elevation of paertures in MEPe increases
                                    ;     ;X(SGI) component.
    ;; e_array is a 2-D array [ azch, 3 ]
    """

    dtor = np.pi / 180.

    e_array = np.array([
     +1 * np.sin(theta_array*dtor),
     -1 * np.cos(theta_array*dtor) * np.sin(phi_array*dtor),
     -1 * np.cos(theta_array*dtor) * np.cos(phi_array*dtor)
     ])

    if fluxdir:
        e_array *= -1.  # ;;Flip the direction

    return e_array



"""Print the version number for the current installation."""
import logging


def version():
    """
    Print the pyspedas version number.

    Returns
    -------
    None.

    """
    import pkg_resources
    ver = pkg_resources.get_distribution("pyspedas").version
    logging.info("pyspedas version: " + ver)



import os
import unittest
from pyspedas.utilities.data_exists import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_downloadonly(self):
        files = pyspedas.polar.efi(downloadonly=True)
        self.assertTrue(os.path.exists(files[0]))

    def test_load_mfe_data(self):
        mfe_vars = pyspedas.polar.mfe(trange=['2003-10-28', '2003-10-29'], get_support_data=True)
        self.assertTrue(data_exists('B_GSE'))
        self.assertTrue(data_exists('B_GSM'))

    def test_load_efi_data(self):
        efi_vars = pyspedas.polar.efi(time_clip=True)
        self.assertTrue(data_exists('ESPIN'))

    def test_load_pwi_data(self):
        pwi_vars = pyspedas.polar.pwi()
        self.assertTrue(data_exists('Fce'))

    def test_load_hydra_data(self):
        hydra_vars = pyspedas.polar.hydra()
        self.assertTrue(data_exists('ELE_DENSITY'))

    def test_load_tide_data(self):
        tide_vars = pyspedas.polar.tide()
        self.assertTrue(data_exists('total_den'))
        self.assertTrue(data_exists('total_v'))
        self.assertTrue(data_exists('total_t'))

    def test_load_timas_data(self):
        timas_vars = pyspedas.polar.timas()
        self.assertTrue(data_exists('Density_H'))

    def test_load_cammice_data(self):
        cammice_vars = pyspedas.polar.cammice()
        self.assertTrue(data_exists('Protons'))

    def test_load_ceppad_data(self):
        ceppad_vars = pyspedas.polar.ceppad()
        self.assertTrue(data_exists('IPS_10_ERR'))

    def test_load_uvi_data(self):
        uvi_vars = pyspedas.polar.uvi()
        self.assertTrue(data_exists('IMAGE_DATA'))

    def test_load_pixie_data(self):
        pixie_vars = pyspedas.polar.pixie()
        self.assertTrue(data_exists('TXF_HIGH'))

    def test_load_vis_data(self):
        vis_vars = pyspedas.polar.vis(notplot=True)
        self.assertTrue('Image_Counts' in vis_vars)

    def test_load_orbit_data(self):
        orbit_vars = pyspedas.polar.orbit()
        self.assertTrue(data_exists('AVG_SPIN_RATE'))
        self.assertTrue(data_exists('SPIN_PHASE'))


if __name__ == '__main__':
    unittest.main()



import logging
from pytplot import get_data


def cotrans_get_coord(name):
    '''
    This function returns the coordinate system of a tplot variable

    Parameters:
        name: str
            name of the tplot variable

    Notes:
        The coordinate system is stored in the variable's metadata at:
            metadata['data_att']['coord_sys']

        See cotrans_set_coord to update the coordinate system

    Returns:
        Coordinate system of the tplot variable 
        or 
        None if the coordinate system isn't set
    '''

    metadata = get_data(name, metadata=True)
    if metadata is None:
        return None

    if metadata.get('data_att'):
        if metadata['data_att'].get('coord_sys'):
            return metadata['data_att']['coord_sys']

    logging.error('Coordinate system not found: ' + name)
    return None



import os

CONFIG = {
    'local_data_dir': 'psp_data/',
    'remote_data_dir': 'https://spdf.gsfc.nasa.gov/pub/data/psp/',
    'sweap_remote_data_dir': 'http://sweap.cfa.harvard.edu/data/sci/',
    'fields_remote_data_dir': 'https://sprg.ssl.berkeley.edu/data/spp/data/sci/'
}

# override local data directory with environment variables
if os.environ.get('SPEDAS_DATA_DIR'):
    CONFIG['local_data_dir'] = os.sep.join([os.environ['SPEDAS_DATA_DIR'], 'psp'])

if os.environ.get('PSP_DATA_DIR'):
    CONFIG['local_data_dir'] = os.environ['PSP_DATA_DIR']


import unittest
from pyspedas.hapi.hapi import hapi
from pyspedas.utilities.data_exists import data_exists


class HAPITests(unittest.TestCase):
    def test_print_servers(self):
        hapi(trange=['2003-10-20', '2003-11-30'])

    def test_print_catalog(self):
        hapi(server='https://cdaweb.gsfc.nasa.gov/hapi', catalog=True)

    def test_dataset_not_specified(self):
        # dataset not specified
        h_vars = hapi(trange=['2003-10-20', '2003-11-30'],
                      server='https://cdaweb.gsfc.nasa.gov/hapi')

    def test_trange_not_specified(self):
        # trange not specified
        h_vars = hapi(dataset='OMNI_HRO2_1MIN',
                      server='https://cdaweb.gsfc.nasa.gov/hapi')

    def test_cdaweb_mms_spec(self):
        h_vars = hapi(trange=['2019-10-16', '2019-10-17'],
                      server='https://cdaweb.gsfc.nasa.gov/hapi',
                      dataset='MMS4_EDP_SRVY_L2_HFESP')

    def test_cdaweb_omni(self):
        h_vars = hapi(trange=['2003-10-20', '2003-11-30'],
                      server='https://cdaweb.gsfc.nasa.gov/hapi',
                      dataset='OMNI_HRO2_1MIN')
        self.assertTrue(data_exists('BX_GSE'))
        self.assertTrue(data_exists('BY_GSE'))
        self.assertTrue(data_exists('BZ_GSE'))


if __name__ == '__main__':
    unittest.main()



import os
import unittest
from pyspedas.utilities.data_exists import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_load_rep_data(self):
        rep_vars = pyspedas.csswe.reptile(time_clip=True)
        self.assertTrue(data_exists('E1flux'))

    def test_load_notplot(self):
        rep_vars = pyspedas.csswe.reptile(notplot=True)
        self.assertTrue('E1flux' in rep_vars)

    def test_downloadonly(self):
        files = pyspedas.csswe.reptile(downloadonly=True, trange=['2014-2-15', '2014-2-16'])
        self.assertTrue(os.path.exists(files[0]))


if __name__ == '__main__':
    unittest.main()



from .load import load
from pyspedas.rbsp.rbspice_lib.rbsp_load_rbspice_read import rbsp_load_rbspice_read
from pyspedas.rbsp.rbspice_lib.rbsp_rbspice_omni import rbsp_rbspice_omni
from pyspedas.rbsp.rbspice_lib.rbsp_rbspice_spin_avg import rbsp_rbspice_spin_avg
from pyspedas.utilities.datasets import find_datasets


def emfisis(trange=['2018-11-5', '2018-11-6'], 
        probe='a',
        datatype='magnetometer', 
        level='l3',
        cadence='4sec', # for EMFISIS mag data
        coord='sm', # for EMFISIS mag 
        wavetype='waveform', # for EMFISIS waveform data
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Electric and Magnetic Field Instrument Suite and Integrated Science (EMFISIS) instrument

    For information on the EMFISIS data products, see:
        https://emfisis.physics.uiowa.edu/data/level_descriptions
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str or list of str
            Spacecraft probe name ('a' or 'b'); default: a

        cadence: str
            Data cadence (default: 4sec); other options: '1sec', 'hires'

        coord: str
            Data coordinate system (default: sm)

        level: str
            Data level; options: 'l1', 'l2', 'l3', l4'

        datatype: str
            Data type; valid options:
            Level 1:
                'magnetometer'
                'hfr'
                'housekeeping'
                'sc-hk'
                'spaceweather'
                'wfr'
                'wna'

            Level 2:
                'magnetometer'
                'wfr'
                'hfr'
                'housekeeping'

            Level 3:
                'magnetometer'

            Level 4:
                'density'
                'wna-survey'

        wavetype: str
            Type of level 2 waveform data; valid options:
                For WFR data:
                'waveform' (default)
                'waveform-continuous-burst'
                'spectral-matrix'
                'spectral-matrix-diagonal'
                'spectral-matrix-diagonal-merged'

                For HFR data:
                'waveform'
                'spectra'
                'spectra-burst'
                'spectra-merged'

            For descriptions of these data, see:
                https://emfisis.physics.uiowa.edu/data/L2_products

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='emfisis', wavetype=wavetype, trange=trange, probe=probe, datatype=datatype, level=level, cadence=cadence, coord=coord, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)


def rbspice(trange=['2018-11-5', '2018-11-6'], 
        probe='a',
        datatype='TOFxEH',
        level='l3',
        suffix='',  
        get_support_data=True,
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Radiation Belt Storm Probes Ion Composition Experiment (RBSPICE) instrument
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str or list of str
            Spacecraft probe name ('a' or 'b'); default: a

        datatype: str
            Data type (default: tofxeh); Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars = load(instrument='rbspice', trange=trange, probe=probe, datatype=datatype, level=level, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    if not isinstance(probe, list):
        probe = [probe]

    for prb in probe:
        # Add energy channel energy values to primary data variable,
        # create variables for individual telescopes, and set appropriate tplot options
        rbsp_load_rbspice_read(level=level, probe=prb, datatype=datatype)

        # Calculate omni-directional variable
        omni_vars = rbsp_rbspice_omni(probe=prb, datatype=datatype, level=level)
        if omni_vars:
            tvars.extend(omni_vars)

        # Calculate spin-averaged variable
        sp_avg_vars = rbsp_rbspice_spin_avg(probe=prb, datatype=datatype, level=level)
        if omni_vars:
            tvars.extend(sp_avg_vars)

    return tvars


def efw(trange=['2015-11-5', '2015-11-6'], 
        probe='a',
        datatype='spec', 
        level='l3',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Electric Field and Waves Suite (EFW)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str or list of str
            Spacecraft probe name ('a' or 'b'); default: a

        datatype: str
            Data type; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='efw', trange=trange, probe=probe, datatype=datatype, level=level, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)


def mageis(trange=['2015-11-5', '2015-11-6'], 
        probe='a',
        datatype='', 
        level='l3',
        rel='rel04',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Energetic Particle, Composition, and Thermal Plasma Suite (ECT)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str or list of str
            Spacecraft probe name ('a' or 'b'); default: a

        datatype: str
            Data type; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='mageis', rel=rel, trange=trange, probe=probe, datatype=datatype, level=level, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)


def hope(trange=['2015-11-5', '2015-11-6'], 
        probe='a',
        datatype='moments', 
        level='l3',
        rel='rel04',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Energetic Particle, Composition, and Thermal Plasma Suite (ECT)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str or list of str
            Spacecraft probe name ('a' or 'b'); default: a

        datatype: str
            Data type; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='hope', rel=rel, trange=trange, probe=probe, datatype=datatype, level=level, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)


def rept(trange=['2015-11-5', '2015-11-6'], 
        probe='a',
        datatype='', 
        level='l3',
        rel='rel03',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Energetic Particle, Composition, and Thermal Plasma Suite (ECT)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str or list of str
            Spacecraft probe name ('a' or 'b'); default: a

        datatype: str
            Data type; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='rept', rel=rel, trange=trange, probe=probe, datatype=datatype, level=level, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)


def rps(trange=['2015-11-5', '2015-11-6'], 
        probe='a',
        datatype='rps-1min', 
        level='l2',
        suffix='',  
        get_support_data=True, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Relativistic Proton Spectrometer (RPS)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str or list of str
            Spacecraft probe name ('a' or 'b'); default: a

        datatype: str
            Data type; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='rps', trange=trange, probe=probe, datatype=datatype, level=level, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)


def datasets(instrument=None, label=True):
    return find_datasets(mission='Van Allen Probes (RBSP)', instrument=instrument, label=label)




import numpy as np

# use nansum from bottleneck if it's installed, otherwise use the numpy one
try:
    import bottleneck as bn
    nansum = bn.nansum
except ImportError:
    nansum = np.nansum


def mms_pgs_make_phi_spec(data_in, resolution=32):
    """

    """
    data = data_in.copy()
    n_phi = resolution

    # zero inactive bins to ensure areas with no data are represented as NaN
    zero_bins = np.argwhere(data['bins'] == 0)
    if zero_bins.size != 0:
        for item in zero_bins:
            data['data'][item[0], item[1]] = 0.0

    ave = np.zeros(n_phi)
    bin_size = 360.0/n_phi
    outbins = np.arange(0, 361, bin_size)

    phi_flat = data['phi'].flatten()
    data_flat = data['data'].flatten()
    bins_flat = data['bins'].flatten()

    for bin_idx in range(0, len(outbins)-1):
        this_bin = np.argwhere((phi_flat >= outbins[bin_idx]) & (phi_flat < outbins[bin_idx+1]))
        if len(this_bin) > 0:
            bins = nansum(bins_flat[this_bin])
            if bins != 0.0:
                ave[bin_idx] += nansum(data_flat[this_bin])/bins
    
    y = outbins[0:n_phi]+0.5*(outbins[1::]-outbins[0:n_phi])

    return (y, ave)




import cdflib
from pytplot import options, ylim, get_data

from ..load import load


def mepe(trange=['2017-03-27', '2017-03-28'],
         datatype='omniflux',
         level='l2',
         suffix='',
         get_support_data=False,
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         uname=None,
         passwd=None,
         time_clip=False,
         ror=True):
    """
    This function loads data from the MEP-e experiment from the Arase mission

    Parameters:
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        level: str
            Data level; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default,
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

        ror: bool
            If set, print PI info and rules of the road

    Returns:
        List of tplot variables created.

    """
    initial_notplot_flag = False
    if notplot:
        initial_notplot_flag = True

    if level == 'l3':
        datatype = '3dflux'

    file_res = 3600. * 24
    prefix = 'erg_mepe_'+level + '_' + datatype + '_'
    pathformat = 'satellite/erg/mepe/'+level+'/'+datatype + \
        '/%Y/%m/erg_mepe_'+level+'_'+datatype+'_%Y%m%d_v??_??.cdf'

    loaded_data = load(pathformat=pathformat, trange=trange, level=level, datatype=datatype, file_res=file_res, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                       varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd)

    if (len(loaded_data) > 0) and ror:

        try:
            if isinstance(loaded_data, list):
                if downloadonly:
                    cdf_file = cdflib.CDF(loaded_data[-1])
                    gatt = cdf_file.globalattsget()
                else:
                    gatt = get_data(loaded_data[-1], metadata=True)['CDF']['GATT']
            elif isinstance(loaded_data, dict):
                gatt = loaded_data[list(loaded_data.keys())[-1]]['CDF']['GATT']

            # --- print PI info and rules of the road

            print(' ')
            print(
                '**************************************************************************')
            print(gatt["LOGICAL_SOURCE_DESCRIPTION"])
            print('')
            print('PI: ', gatt['PI_NAME'])
            print("Affiliation: "+gatt["PI_AFFILIATION"])
            print('')
            print('- The rules of the road (RoR) common to the ERG project:')
            print(
                '      https://ergsc.isee.nagoya-u.ac.jp/data_info/rules_of_the_road.shtml.en')
            print(
                '- RoR for MEP-e data:  https://ergsc.isee.nagoya-u.ac.jp/mw/index.php/ErgSat/Mepe')
            print('')
            print('Contact: erg_mep_info at isee.nagoya-u.ac.jp')
            print(
                '**************************************************************************')
        except:
            print('printing PI info and rules of the road was failed')

    if initial_notplot_flag or downloadonly:
        return loaded_data

    if datatype == 'omniflux':
        # set spectrogram plot option
        options(prefix + 'FEDO' + suffix, 'Spec', 1)
        # set y axis to logscale
        options(prefix + 'FEDO' + suffix, 'ylog', 1)
        # set ytitle
        options(prefix + 'FEDO' + suffix, 'ytitle', 'ERG\nMEP-e\nFEDO\nEnergy')
        # set ysubtitle
        options(prefix + 'FEDO' + suffix, 'ysubtitle', '[keV]')
        # set ylim
        ylim(prefix + 'FEDO' + suffix, 6., 100.)
        # set z axis to logscale
        options(prefix + 'FEDO' + suffix, 'zlog', 1)
        # set ztitle
        options(prefix + 'FEDO' + suffix, 'ztitle', '[/s-cm^{2}-sr-keV]')
        # change colormap option
        options(prefix + 'FEDO' + suffix, 'Colormap', 'jet')
    elif (datatype == '3dflux') and (level == 'l2'):
        # set spectrogram plot option
        options(prefix + 'FEDU' + suffix, 'Spec', 1)
        options(prefix + 'FEDU_n' + suffix, 'Spec', 1)
        options(prefix + 'FEEDU' + suffix, 'Spec', 1)
        options(prefix + 'count_raw' + suffix, 'Spec', 1)

        # set y axis to logscale
        options(prefix + 'FEDU' + suffix, 'ylog', 1)
        options(prefix + 'FEDU_n' + suffix, 'ylog', 1)
        options(prefix + 'FEEDU' + suffix, 'ylog', 1)
        options(prefix + 'count_raw' + suffix, 'ylog', 1)

        # set ysubtitle
        options(prefix + 'FEDU' + suffix, 'ysubtitle', '[keV]')
        options(prefix + 'FEDU_n' + suffix, 'ysubtitle', '[keV]')
        options(prefix + 'count_raw' + suffix, 'ysubtitle', '[keV]')

        # set ylim
        ylim(prefix + 'FEDU' + suffix, 6., 100.)
        ylim(prefix + 'FEDU_n' + suffix, 6., 100.)
        ylim(prefix + 'count_raw' + suffix, 6., 100.)

        # set z axis to logscale
        options(prefix + 'FEDU' + suffix, 'zlog', 1)
        options(prefix + 'FEDU_n' + suffix, 'zlog', 1)
        options(prefix + 'FEEDU' + suffix, 'zlog', 1)
        options(prefix + 'count_raw' + suffix, 'zlog', 1)

        # set ztitle
        options(prefix + 'FEDU' + suffix, 'ztitle', '[/s-cm^{2}-sr-keV]')
        options(prefix + 'FEDU_n' + suffix, 'ztitle', '[/s-cm^{2}-sr-keV]')

        # change colormap option
        options(prefix + 'FEDU' + suffix, 'Colormap', 'jet')
        options(prefix + 'FEDU_n' + suffix, 'Colormap', 'jet')
        options(prefix + 'FEEDU' + suffix, 'Colormap', 'jet')
        options(prefix + 'count_raw' + suffix, 'Colormap', 'jet')

    return loaded_data



import logging

import numpy as np

from copy import deepcopy

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s',
                    datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)


def erg_convert_flux_units(input_dist, units='flux', relativistic=False):
    """
    ; The following unit names are acceptable for units:
    ;   'flux' 'eflux' 'df' 'df_cm'
    ;
    ;   'df_km' and 'psd' are referred to as 'df'.
    ;
    ; CAUTION!!!
    ; "relativistic" keyword is valid only for electron currently.
    ; Using it for ions just messes up the conversion.
    """

    output_dist = deepcopy(input_dist)

    units_out = units.lower()
    species_lc = input_dist['species'].lower()
    units_in = input_dist['units_name'].lower()

    if units_in == units_out:
        return input_dist

    # Unify some unit notations

    if (units_in == 'df_km') or (units_in == 'psd'):
        units_in = 'df'
    if (units_out == 'df_km') or (units_out == 'psd'):
        units_out = 'df'

    # Get the mass of species (unit: proton mass)
    if species_lc == 'e':
        A = 1.0/1836.0  # e-
    elif species_lc == 'hplus':
        A = 1.0  # H+
    elif species_lc == 'proton':
        A = 1.0  # H+
    elif species_lc == 'he2plus':
        A = 4.0  # He2+
    elif species_lc == 'alpha':
        A = 4.0  # He2+
    elif species_lc == 'heplus':
        A = 4.0  # He+
    elif species_lc == 'oplusplus':
        A = 16.0  # O++
    elif species_lc == 'oplus':
        A = 16.0  # O+
    elif species_lc == 'o2plus':
        A = 32.0  # (O2)+

    """
    ;; Scaling factor between df (s^3/km^6) and flux (#/eV/s/str/cm2).
    ;; Note that the notation is in such a way that the conversion is
    ;; done below as:
    ;; f [s3/km6] = j [#/eV/s/str/cm2] / K [eV] * flux_to_df
    ;;                                 for non-relativistic cases
    ;;
    ;; f [(c/MeV/cm)^3] = j [#/eV/s/str/cm2] * flux_to_df / K [eV]
    ;;                 for relativisitc electron cases with keyword relativistic on
    ;; 
    ;; !!CAUTION!! MEPs data should be converted to (#/eV/s/str/cm2) 
    ;; when one uses the mep_part_products libraries.
    ;;
    ;; !!CAUTION2!!
    ;; Keyword "relativistic" is valid for only electrons.
    ;; DO NOT USE it for ions. 
    """

    flux_to_df = A**2.0 * 0.5447 * 1e6

    if relativistic:
        # Conversion here is based on those adopted by Hilmer+JGR,2000.

        mc2 = 5.10999e-1  # Electron rest energy [MeV]
        ene = deepcopy(input_dist['energy'])  # [eV]
        MeV_ene = ene * 1e-6  # [MeV]
        p2c2 = MeV_ene * (MeV_ene + 2.0 * mc2)  # [MeV^2]

        """
        ;; f [(c/MeV/cm)^3]
        ;;     = j [#/eV/s/str/cm2] * 1d+3 / p2c2 * 1.66d-10 * 200.3 
        ;; 1d+3 is to convert input flux values to [#/keV/s/sr/cm2]. 
        ;; The multiplication of energy [eV] is to be consistent
        ;; with the conversion below. 
        """
        flux_to_df = 1.0e+3 / p2c2 * 1.66e-10 * 200.3 * ene

    # factor between km^6 and cm^6 for df
    cm_to_km = 1e+30

    """
    ;; Calculation will be kept simple and stable as possible by 
    ;; pre-determining the final exponent of each scaling factor 
    ;; rather than multiplying by all applicable in/out factors
    ;; these exponents should always be integers!
    ;;    [energy, flux_to_df, cm_to_km]
    """
    exp_in = [0, 0, 0]
    exp_out = [0, 0, 0]

    """
    ;; get input/output scaling exponents  
    ;; All conversions are done via energy flux (eflux).
    """

    if units_in == 'flux':
        exp_in = [1, 0, 0]
    elif units_in == 'df':
        exp_in = [2, -1, 0]
    elif units_in == 'df_cm':
        exp_in = [2, -1, 1]

    if units_out == 'flux':
        exp_out = [-1, 0, 0]
    elif units_out == 'df':
        exp_out = [-2, 1, 0]
    elif units_out == 'df_cm':
        exp_out = [-2, 1, -1]

    exp = np.array(exp_in) + np.array(exp_out)

    """
    ;; Ensure everything is double prec first for numerical stability
    ;;  -target field won't be mutated since it's part of a structure
    """

    output_dist['data'] = input_dist['data'] * input_dist['energy']**exp[0] * \
        (flux_to_df ** exp[1] * cm_to_km ** exp[2])

    output_dist['units_name'] = units_out

    return output_dist



from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.analysis.time_clip import time_clip as tclip
from pytplot import cdf_to_tplot

from .config import CONFIG


def load(trange=['1983-02-16', '1983-02-17'], 
         instrument='mag',
         datatype='',
         suffix='', 
         get_support_data=False, 
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False):
    """
    This function loads data from the DE2 mission; this function is not meant 
    to be called directly; instead, see the wrappers:

        pyspedas.de2.mag
        pyspedas.de2.nacs
        pyspedas.de2.rpa
        pyspedas.de2.fpi
        pyspedas.de2.idm
        pyspedas.de2.wats
        pyspedas.de2.vefi
        pyspedas.de2.lang

    """

    if instrument == 'mag':
        pathformat = 'magnetic_electric_fields_vefi_magb/'+datatype+'_vefimagb_cdaweb/%Y/de2_'+datatype+'_vefimagb_%Y%m%d_v??.cdf'
    elif instrument == 'nacs':
        pathformat = 'neutral_gas_nacs/'+datatype+'_'+instrument+'_cdaweb/%Y/de2_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
    elif instrument == 'rpa':
        pathformat = 'plasma_rpa/'+datatype+'_cdaweb/%Y/de2_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
    elif instrument == 'fpi':
        pathformat = 'neutral_gas_fpi/de2_neutral8s_fpi/%Y/de2_neutral'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
    elif instrument == 'idm':
        pathformat = 'plasma_idm/vion250ms_cdaweb/%Y/de2_vion'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
    elif instrument == 'wats':
        pathformat = 'neutral_gas_wats/wind2s_wats_cdaweb/%Y/de2_wind'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
    elif instrument == 'vefi':
        pathformat = 'electric_fields_vefi/'+datatype+'_vefi_cdaweb/%Y/de2_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
    elif instrument == 'lang':
        pathformat = 'plasma_lang/plasma500ms_lang_cdaweb/%Y/de2_plasma'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange)

    out_files = []

    files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG['local_data_dir'], no_download=no_update)
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(out_files, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, notplot=notplot)
    
    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars

    


import logging
import numpy as np
from fnmatch import fnmatch
from pytplot import get_data, store_data, options
from pyspedas import time_datetime


def mms_fpi_make_compressionlossbars(tname, lossy=False):
    """
    Creates FPI's compressionloss flag bars

    Parameters
    -----------
        tname: str
            Tplot variable name of DIS or DES compressionloss data

        lossy: bool
            The value for lossy compression (use this keyword only for special case)

    Flag
    -----------
        0: Lossless compression
        1: Lossy compression
      In old files (v2.1.0 or older until the end of Phase 1A)
        1: Lossless compression
        3: Lossy compression

    Notes
    -----------
        In cases when data had been lossy compressed, some artifacts may appear in the data due to the compression.
        Since all of fast survey data are lossy compressed, it is not necessary to make this bar for fast survey data.

    Returns
    ------------
        List of the tplot variables created.

    """

    if fnmatch(tname, 'mms?_dis*'):
        instrument = 'DIS'
    elif fnmatch(tname, 'mms?_des*'):
        instrument = 'DES'
    else:
        logging.error('Unable to determine instrument from variable name.')
        return

    if instrument == 'DES':
        colors = 'red'
    else:
        colors = 'blue'

    if fnmatch(tname, '*_fast*'):
        logging.info('All fast survey data are lossy compressed, so there is no need to create this bar.')
        return
    elif fnmatch(tname, '*_brst*'):
        data_rate = 'Brst'
    else:
        logging.error('Unable to determine data rate from variable name.')
        return

    data = get_data(tname, dt=True)
    metadata = get_data(tname, metadata=True)

    if data is None:
        logging.error('Problem reading the variable: ' + tname)
        return

    flagline = np.zeros(len(data.times))

    if not lossy:
        file_id = metadata['CDF']['GATT']['Logical_file_id']
        version = file_id.split('_v')[1].split('.')
        if version[0] == '2':
            if version[1] == '1':
                if data.times[0] < time_datetime('2016-04-01'):
                    lossy = 3
                else:
                    lossy = 1
            else:
                if float(version[1]) > 1:
                    lossy = 1
                else:
                    lossy = 3
        else:
            if float(version[0]) > 2:
                lossy = 1
            else:
                lossy = 3

        for j in range(len(data.times)):
            if data.y[j] != lossy:
                flagline[j] = np.nan
            else:
                flagline[j] = 0.5

        store_data(tname + '_flagbars', data={'x': data.times, 'y': flagline})
        options(tname + '_flagbars', 'yrange', [0, 1])
        options(tname + '_flagbars', 'panel_size', 0.2)
        options(tname + '_flagbars', 'symbols', True)
        options(tname + '_flagbars', 'markers', 's')
        options(tname + '_flagbars', 'thick', 4)
        options(tname + '_flagbars', 'border', False)

        return [tname + '_flagbars']



import unittest
import pyspedas
from pyspedas import sosmag_load
from pyspedas.utilities.data_exists import data_exists

pyspedas.sosmag.load.sosmag_parameters['print_messages'] = True


class SOSMAG_Tests(unittest.TestCase):
    def test_mag(self):
        t_ok, var_names = sosmag_load(trange=['2021-01-01 02:00', '2021-01-01 03:00'])
        self.assertTrue(data_exists('sosmag_b_gse'))
        self.assertTrue(data_exists('sosmag_position'))

    def test_1m(self):
        t_ok, var_names = sosmag_load(datatype='1m', trange=['2021-01-01 02:00', '2021-01-01 03:00'])
        self.assertTrue(data_exists('sosmag_1m_b_gse'))
        self.assertTrue(data_exists('sosmag_1m_position'))

    def test_invalid_user(self):
        pyspedas.sosmag.load.sosmag_parameters['username'] = 'not_valid'
        try:
            t_ok, var_names = sosmag_load(datatype='1m', trange=['2021-01-01 02:00', '2021-01-01 03:00'])
        except:
            pass
        pyspedas.sosmag.load.sosmag_parameters['username'] = 'spedas'

    def test_invalid_cookies(self):
        authenticated, auth_cookie = pyspedas.sosmag.load.sosmag_get_auth_cookie()
        # shouldn't work
        invalid = pyspedas.sosmag.load.sosmag_get_session('')
        invalid = pyspedas.sosmag.load.sosmag_get_capabilities('', '')
        invalid = pyspedas.sosmag.load.sosmag_get_data('', '')
        invalid = pyspedas.sosmag.load.sosmag_to_tplot([1], '', '')
        invalid = pyspedas.sosmag.load.sosmag_to_tplot(None, '', '')


if __name__ == '__main__':
    unittest.main()



import numpy as np

from astropy.coordinates import cartesian_to_spherical


def get_mepi_flux_angle_in_sga(looking_dir=False):

    #;; --- starts from -Z(SGI) axis and increases toward -Y(SGI) axis.
    azi_0 = 11.25  # ;;[deg] offset of channel 0 (anode 00) from -Z(SGI) axis
    d_azi = 22.5   # ;;[deg] angula interval between the centers of neighboring ch

    phi_array = azi_0 + (np.arange(16)) * d_azi

    #;; Inclination of apertures, based on visual inspection
    theta_array = np.full(shape=(16), fill_value=1.6)

    """
    ;; unit vector of looking direction 
    ;;   e_array is a 2-D array [azch, 3(x,y,z)] in SGA/SGI 
    """
    dtor = np.pi / 180.

    e_array = np.array([
     -1 * np.sin(theta_array*dtor),
     -1 * np.cos(theta_array*dtor) * np.sin(phi_array*dtor),
     -1 * np.cos(theta_array*dtor) * np.cos(phi_array*dtor)
     ])

    if not looking_dir:
        e_array *= -1.  # ;;Flip the directions to be flux dirs

    spherical_data = cartesian_to_spherical(
        x=e_array[0],
        y=e_array[1],
        z=e_array[2]
    )

    rtod = 180. / np.pi
    elev_array = spherical_data[1].value * rtod
    phi_array = spherical_data[2].value * rtod

    anglarr = np.zeros(shape=(2, 3, 16))  # ;[ elev/phi, min/cnt/max, apd_no ]

    anglarr[0] = elev_array
    anglarr[1] = phi_array

    return anglarr



import logging
from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.analysis.time_clip import time_clip as tclip
from pytplot import cdf_to_tplot

from .config import CONFIG


def load(trange=['2006-06-01', '2006-06-02'], 
         instrument='celias',
         datatype='pm_5min',
         suffix='', 
         get_support_data=False, 
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False):
    """
    This function loads data from the SOHO mission; this function is not meant 
    to be called directly; instead, see the wrappers:

        pyspedas.soho.celias
        pyspedas.soho.cosp
        pyspedas.soho.erne
        pyspedas.soho.orbit

    """
    res = 24 * 3600.

    if instrument == 'celias':
        pathformat = instrument+'/'+datatype+'/%Y/soho_'+instrument+'-'+datatype+'_%Y%m%d_v??.cdf'
    elif instrument == 'costep':
        pathformat = instrument+'/'+datatype+'/%Y/soho_'+instrument+'-'+datatype+'_%Y0101_v??.??.cdf'
        res = 24 * 3600. * 366
    elif instrument == 'erne':
        pathformat = instrument+'/'+datatype+'/%Y/soho_'+instrument+'-'+datatype+'_%Y%m%d_v??.cdf'
    elif instrument == 'orbit':
        if datatype not in ['pre_or', 'def_or', 'def_at']:
            logging.error('Invalod datatype: ' + datatype)
            return
        datatype_fn = datatype.split('_')[1] + '_' +datatype.split('_')[0]
        pathformat = instrument+'/'+datatype+'/cdf/%Y/so_'+datatype_fn+'_%Y%m%d_v??.cdf'

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange, res=res)

    out_files = []

    files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG['local_data_dir'], no_download=no_update)
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(out_files, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, notplot=notplot)
    
    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars

    


import os
import unittest
from pyspedas.utilities.data_exists import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_load_pws_data(self):
        out_vars = pyspedas.akebono.pws(time_clip=True)
        self.assertTrue(data_exists('akb_pws_RX1'))
        self.assertTrue(data_exists('akb_pws_RX2'))

    def test_load_rdm_data(self):
        out_vars = pyspedas.akebono.rdm()
        self.assertTrue(data_exists('akb_L'))
        self.assertTrue(data_exists('akb_MLT'))
        self.assertTrue(data_exists('akb_rdm_FEIO'))

    def test_load_orb_data(self):
        out_vars = pyspedas.akebono.orb()
        self.assertTrue(data_exists('akb_orb_geo'))
        self.assertTrue(data_exists('akb_orb_gdlat'))
        self.assertTrue(data_exists('akb_orb_gdlon'))

    def test_load_notplot(self):
        out_vars = pyspedas.akebono.pws(notplot=True)
        self.assertTrue('akb_pws_RX1' in out_vars)

    def test_downloadonly(self):
        files = pyspedas.akebono.pws(downloadonly=True, trange=['2012-10-01', '2012-10-02'])
        self.assertTrue(os.path.exists(files[0]))


if __name__ == '__main__':
    unittest.main()

    


import logging
import numpy as np
from pytplot import get_data, store_data, options
from pyspedas.rbsp.rbspice_lib.rbsp_rbspice_pad_spinavg import rbsp_rbspice_pad_spinavg


def rbsp_rbspice_pad(probe='a', datatype='TOFxEH', level='l3', energy=[0, 1000], bin_size=15, scopes=None):
    """
    Calculate pitch angle distributions using data from the
    RBSP Radiation Belt Storm Probes Ion Composition Experiment (RBSPICE)
    
    Parameters
    ----------
    probe : str
        RBSP spacecraft indicator [Options: 'a' (default), 'b']
    datatype : str
        desired data type [Options: 'TOFxEH' (default), 'TOFxEnonH']
    level : str
        data level ['l1','l2','l3' (default),'l3pap']
    energy : list
        user-defined energy range to include in the calculation in keV [default = [0,1000]]
    bin_size : float
        desired size of the pitch angle bins in degrees [default = 15]
    scopes : list
        string array of telescopes to be included in PAD [0-5, default is all]
    
    Returns
    -------
    Tplot variables created
    """
    if datatype == 'TOFxEH':
        species = 'proton'
    elif datatype == 'TOFxEnonH':
        species = ['helium', 'oxygen']
    elif datatype == 'TOFxPHHHELT':
        species = ['proton', 'oxygen']

    if not isinstance(species, list):
        species = [species]

    if level != 'l1':
        units_label = '1/(cm^2-sr-s-keV)'
    else:
        units_label = 'counts/s'
    if not energy:
        energy = [0, 1000]
    if not bin_size:
        bin_size = 15.
    if not scopes:
        scopes = [0, 1, 2, 3, 4, 5]
    
    prefix = 'rbsp'+probe+'_rbspice_'+level+'_'+datatype+'_'
    
    if energy[0] > energy[1]:
        logging.error('Low energy must be given first, then high energy in "energy" keyword')
        return
    
    # set up the number of pa bins to create
    bin_size = float(bin_size)
    n_pabins = int(180./bin_size)
    pa_bins = 180.*np.arange(n_pabins+1)/n_pabins
    pa_label = 180.*np.arange(n_pabins)/n_pabins+bin_size/2.
    
    logging.info('Num PA bins: ' + str(n_pabins))
    logging.info('PA bins: ' + str(pa_bins))
 
    # check to make sure the data exist
    d = get_data(prefix + 'Alpha')
    if d is None:
        logging.error('No '+datatype+' data is currently loaded for probe rbsp-'+probe+' for the selected time period')
        return

    logging.info('Calculating RBSPICE pitch angle distribution..')
    out = []

    for ion_type_idx in range(len(species)):
        # get pitch angle data (all telescopes in single variable)
        d_pa = get_data(prefix + 'Alpha')
        pa_file = np.zeros((len(d_pa.times), len(scopes))) # time steps, look direction
        for aa in range(len(scopes)):
            pa_file[:, scopes[aa]] = d_pa.y[:, scopes[aa]]

        pa_flux = np.zeros((len(d_pa.times), n_pabins, len(scopes)))
        pa_flux_nans = np.argwhere(pa_flux == 0)
        if len(pa_flux_nans) > 0:
            pa_flux[pa_flux_nans] = np.nan
        pa_num_in_bin = np.zeros((len(d_pa.times), n_pabins, len(scopes)))

        for qq in range(len(species)):
            # get flux data (all telescopes in single variable)
            d_flux = get_data(prefix + species[qq])
            d_flux_t0 = get_data(prefix + species[qq] + '_T0')

            logging.info(prefix + species[qq])
            flux_file = np.zeros((len(d_flux.times), len(scopes))) # time steps, look direction
            flux_file_nans = np.argwhere(flux_file == 0)
            if len(flux_file_nans) > 0:
                flux_file[flux_file_nans] = np.nan
            new_pa_flux = np.zeros((len(d_flux.times), n_pabins, len(scopes)))          # the average for each bin

            # get energy range of interest
            e = d_flux_t0.v
            indx = np.argwhere((e < energy[1]) & (e > energy[0]))
            energy_count = len(indx[0])

            if energy_count == 0:
                logging.warning('Energy range selected is not covered by the detector for ' + datatype + ' ' + species[ion_type_idx])
                continue

            for t in range(len(scopes)):
                # Loop through each time step and get:
                # 1.  the total flux for the energy range of interest for each detector
                # 2.  flux in each pa bin
                for i in range(len(d_flux.times)): # loop through time
                    flux_file[i, t] = np.nansum(d_flux.y[i, indx, scopes[t]])  # start with lowest energy
                    for j in range(n_pabins): # loop through pa bins
                        if (pa_file[i, t] > pa_bins[j]) and (pa_file[i,t] < pa_bins[j+1]):
                            if not np.isfinite(pa_flux[i, j, t]):
                                pa_flux[i, j, t] = flux_file[i, t]
                            else:
                                pa_flux[i, j, t] = pa_flux[i, j, t] + flux_file[i, t]
                            pa_num_in_bin[i, j, t] += 1.0

                # loop over time
                for i in range(len(pa_flux[:, 0, 0])):
                    # loop over bins
                    for bin_idx in range(len(pa_flux[i, :, 0])):
                        if pa_num_in_bin[i, bin_idx, t] != 0.0:
                            new_pa_flux[i, bin_idx, t] = pa_flux[i, bin_idx, t]/pa_num_in_bin[i, bin_idx, t]
                        else:
                            new_pa_flux[i, bin_idx, t] = np.nan

            en_range_string = str(energy[0]).replace('.', '') + '-' + str(energy[1]).replace('.', '') + 'keV'
            if len(scopes) == 6:
                new_name = prefix+species[qq]+'_omni_'+en_range_string+'_pad'
                new_omni_pa_flux = np.zeros((len(new_pa_flux[:, 0, 0]),len(new_pa_flux[0, :, 0])))
                for ii in range(len(new_pa_flux[:, 0, 0])):
                    for jj in range(len(new_pa_flux[0, :, 0])):
                        new_omni_pa_flux[ii, jj] = np.nanmean(new_pa_flux[ii, jj, :])
                store_data(new_name, data={'x': d_flux.times, 'y': new_omni_pa_flux, 'v': pa_label})
                options(new_name, 'yrange', [0, 180])
                options(new_name, 'spec', True)
                options(new_name, 'zlog', True)
                options(new_name, 'ytitle', 'rbsp-'+probe+'\nrbspice\n'+species[ion_type_idx]+'\nomni')
                options(new_name, 'ysubtitle', en_range_string+'\nPA [Deg]')
                options(new_name, 'ztitle', units_label)
                out.append(new_name)
            else:
                new_name = []
                for ii in range(len(scopes)):
                    new_name.append(prefix+species[qq]+'_T'+str(scopes[ii])+'_'+en_range_string+'_pad')
                    store_data(new_name[ii], data={'x': d_flux.times, 'y': new_pa_flux[:, :, ii], 'v': pa_label})
                    options(new_name[ii], 'yrange', [0, 180])
                    options(new_name[ii], 'spec', True)
                    options(new_name[ii], 'zlog', True)
                    options(new_name[ii], 'ytitle', 'rbsp-'+probe+'\nrbspice\n' +species[ion_type_idx]+'\nT'+str(scopes[t]))
                    options(new_name[ii], 'ysubtitle', en_range_string + '\nPA [Deg]')
                    options(new_name[ii], 'ztitle', units_label)
                    out.append(new_name[ii])

            # now do the spin average
            sp_vars = rbsp_rbspice_pad_spinavg(probe=probe, datatype=datatype, species=species[ion_type_idx], energy=energy, bin_size=bin_size, scopes=scopes)
            out.extend(sp_vars)
    return out



import os
import unittest
from pyspedas.utilities.data_exists import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_load_cindi_data(self):
        c_vars = pyspedas.cnofs.cindi(time_clip=True)
        self.assertTrue(data_exists('ionVelocityX'))
        self.assertTrue(data_exists('ionVelocityY'))
        self.assertTrue(data_exists('ionVelocityZ'))

    def test_load_plp_data(self):
        l_vars = pyspedas.cnofs.plp()
        self.assertTrue(data_exists('Ni'))

    def test_load_vefi_data(self):
        l_vars = pyspedas.cnofs.vefi()
        self.assertTrue(data_exists('E_meridional'))
        self.assertTrue(data_exists('E_zonal'))

    def test_load_notplot(self):
        c_vars = pyspedas.cnofs.cindi(notplot=True)
        self.assertTrue('ionVelocityX' in c_vars)

    def test_downloadonly(self):
        files = pyspedas.cnofs.cindi(downloadonly=True, trange=['2013-2-15', '2013-2-16'])
        self.assertTrue(os.path.exists(files[0]))


if __name__ == '__main__':
    unittest.main()



import logging
import numpy as np
from pytplot import get_data, store_data, options
from pyspedas import tnames


def rbsp_rbspice_spin_avg(probe='a', datatype='TOFxEH', level='l3'):
    """
    Calculates spin-averaged fluxes for the RBSPICE instrument
    
    Parameters
    ----------
    probe : str
        RBSP spacecraft indicator [Options: 'a' (default), 'b']
    datatype : str
        RBSPICE data type ['TOFxEH' (default),'TOFxEnonH']
    level : str
        data level ['l1','l2','l3' (default),'l3pap']

    Returns
    --------
    Tplot variables created
    """
    if probe is None:
        probe = 'a'
    if datatype is None:
        datatype = 'TOFxEH'
    if level is None:
        level = 'l3'
    if level != 'l1':
        units_label = '1/(cm^2-sr-s-keV)'
    else:
        units_label = 'counts/s'

    prefix = 'rbsp'+probe+'_rbspice_'+level+'_'+datatype+'_'

    spin_nums = get_data(prefix + 'Spin')
    if spin_nums is None:
        return
    spin_starts = np.unique(spin_nums.y, return_index=True)[1][1:]-1

    if datatype == 'TOFxEH':
        species = 'proton'
    elif datatype == 'TOFxEnonH':
        species = ['helium', 'oxygen']
    elif datatype == 'TOFxPHHHELT':
        species = ['proton', 'oxygen']

    if isinstance(species, list):
        var_data = []
        for spc in species:
            var_data.extend(tnames(prefix + spc + '_T?'))
            var_data.extend(tnames(prefix + spc + '_omni'))
    else:
        var_data = tnames(prefix + species + '_T?')
        var_omni = tnames(prefix + species + '_omni')
        var_data.extend(var_omni)

    logging.info('Calculating spin averaged energy spectra..')
    out = []
    zrange = None

    for n in range(len(var_data)):
        if var_data[n] == '':
            logging.error('Error, problem finding the tplot variables to calculate the spin averages')
            return
        else:
            flux_data = get_data(var_data[n])
            if len(flux_data) < 3:
                logging.error('Error, couldn''t find energy table for the flux/cps data variable')
                continue
            if var_data[n][-2:-1] == 'T':
                species = var_data[n][-9:-3]
            elif var_data[n][-4:] == 'omni':
                species = var_data[n][-11:-5]
            if species == 'proton':
                if datatype != 'TOFxPHHHELT':
                    zrange = [5., 1.e5]
                else:
                    zrange = [2.e2, 1.e6]
            elif species == 'helium':
                zrange = [1., 5.e2]
            elif species == 'oxygen':
                if datatype != 'TOFxPHHHELT':
                    zrange = [1., 1.e2]
                else:
                    zrange = [1e1, 1.e4]

            spin_sum_flux = np.zeros((len(spin_starts), len(flux_data.v)))
            current_start = 0
            for spin_idx in range(len(spin_starts)):
                spin_sum_flux[spin_idx, :] = np.nanmean(flux_data.y[current_start:spin_starts[spin_idx]+1, :], axis=0)
                current_start = spin_starts[spin_idx]+1
            sp = '_spin'
            if var_data[n][-4:] == 'omni':
                suffix = ''
            elif var_data[n][-2] == 'T'+str(n):
                suffix = '_T'+str(n)
            else:
                suffix = ''
            store_data(var_data[n]+sp+suffix, data={'x': spin_nums.times[spin_starts], 'y': spin_sum_flux, 'v': flux_data.v})
            options(var_data[n]+sp+suffix, 'ylog', True)
            options(var_data[n]+sp+suffix, 'zlog', True)
            options(var_data[n]+sp+suffix, 'spec', True)
            if zrange is not None:
                options(var_data[n]+sp+suffix, 'zrange', zrange)
            if isinstance(species, list):
                options(var_data[n]+sp+suffix, 'ytitle', 'rbsp'+probe+'\nrbspice\n'+datatype+'\n'+suffix)
            else:
                options(var_data[n]+sp+suffix, 'ytitle', 'rbsp'+probe+'\nrbspice\n'+species+'\n'+suffix)
            options(var_data[n]+sp+suffix, 'ysubtitle', '[keV]')
            options(var_data[n]+sp+suffix, 'ztitle', units_label)
            out.append(var_data[n]+sp+suffix)
    return out



"""
Creates a new tplot variable that has spikes removed.

Notes
-----
Similar to clean_spikes.pro in IDL SPEDAS.

"""
import logging
import numpy as np
import pyspedas
import pytplot
from pyspedas.analysis.subtract_average import subtract_average
from pyspedas.analysis.tsmooth import tsmooth


def clean_spikes(names, nsmooth=10, thresh=0.3, sub_avg=False,
                 new_names=None, suffix=None, overwrite=None):
    """
    Clean spikes from data.

    Parameters
    ----------
    names: str/list of str
        List of pytplot names.
    new_names: str/list of str, optional
        List of new_names for pytplot variables.
        If not given, then a suffix is applied.
    suffix: str, optional
        A suffix to apply. Default is '-avg'.
    overwrite: bool, optional
        Replace the existing tplot name.
    nsmooth: int, optional
        the number of data points for smoothing
    thresh: float, optional
        threshold value
    sub_avg: bool, optional
        if set, subtract the average value of the data
        prior to checking for spikes

    Returns
    -------
    None.

    """
    old_names = pyspedas.tnames(names)

    if len(old_names) < 1:
        logging.error('clean_spikes error: No pytplot names were provided.')
        return

    if suffix is None:
        suffix = '-despike'

    if overwrite is not None:
        n_names = old_names
    elif new_names is None:
        n_names = [s + suffix for s in old_names]
    else:
        n_names = new_names

    if isinstance(n_names, str):
        n_names = [n_names]

    if len(n_names) != len(old_names):
        n_names = [s + suffix for s in old_names]

    for old_idx, old in enumerate(old_names):
        new = n_names[old_idx]
        tmp = new + '_tmp_data'

        # Create new
        if old != new:
            pyspedas.tcopy(old, new)

        # Perform subtract_average or just copy the values
        if sub_avg:
            subtract_average(new, new_names=tmp)
        else:
            pyspedas.tcopy(new, tmp)

        # Find spikes
        tmps = tmp + '-s'
        tsmooth(tmp, new_names=tmps, width=nsmooth)
        ds0 = pytplot.get_data(tmps)  # smoothed out values
        ds = ds0[1]
        dor0 = pytplot.get_data(tmp)  # original values
        d0 = dor0[1]
        dn = d0.copy()  # final values

        dim = dn.shape
        if len(dim) == 1:
            # One dim data.
            for i in range(dim[0]):
                # compare smoothed out values to original values
                if abs(d0[i] - ds[i]) > thresh * abs(ds[i]):
                    dn[i] = np.NaN  # for spikes, set to NaN
        else:
            # More than one dim data.
            for j in range(dim[1]):
                for i in range(dim[0]):
                    # compare smoothed out values to original values
                    if abs(d0[i, j] - ds[i, j]) > thresh * abs(ds[i, j]):
                        dn[i, j] = np.NaN  # for spikes, set to NaN

        # pytplot.data_quants[new] = d
        pytplot.replace_data(new, dn)

        # remove temp data
        del pytplot.data_quants[tmp]
        del pytplot.data_quants[tmps]

        logging.info('clean_spikes was applied to: ' + new)



from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.analysis.time_clip import time_clip as tclip
from pytplot import cdf_to_tplot

from .config import CONFIG

def load(trange=['2009-01-01', '2009-01-02'], 
         instrument='vhm',
         datatype='1min', 
         suffix='', 
         get_support_data=False, 
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False):
    """
    This function is not meant to be called directly; please see the instrument specific wrappers in __init__.py
    """

    if instrument == 'vhm':
        pathformat = 'mag_cdaweb/vhm_'+datatype+'/%Y/uy_'+datatype+'_vhm_%Y%m%d_v??.cdf'
    elif instrument == 'swoops':
        if datatype in ['bai_m0', 'bai_m1', 'bae_m0']:
            pathformat = 'plasma/swoops_cdaweb/'+datatype+'/%Y/uy_'+datatype.split('_')[1]+'_'+datatype.split('_')[0]+'_%Y%m%d_v??.cdf'
        else:
            pathformat = 'plasma/swoops_cdaweb/'+datatype+'/%Y/uy_'+datatype+'_%Y0101_v??.cdf'
    elif instrument == 'swics':
        pathformat = 'plasma/swics_cdaweb/'+datatype+'/%Y/uy_'+datatype.split('_')[1]+'_'+datatype.split('_')[0]+'_%Y%m%d_v??.cdf'
    elif instrument == 'urap':
        pathformat = 'radio/urap_cdaweb/'+datatype+'/%Y/uy_'+datatype.split('_')[1]+'_'+datatype.split('_')[0]+'_%Y%m%d_v??.cdf'
    elif instrument == 'epac':
        if datatype == 'epac_m1':
            pathformat = 'particle/epac_cdaweb/'+datatype+'/%Y/uy_m1_epa_%Y%m%d_v??.cdf'
    elif instrument == 'hiscale':
        pathformat = 'particle/hiscale_cdaweb/'+datatype+'/%Y/uy_'+datatype.split('_')[1]+'_'+datatype.split('_')[0]+'_%Y%m%d_v??.cdf'
    elif instrument == 'cospin':
        pathformat = 'particle/cospin_cdaweb/'+datatype+'/%Y/uy_m0_'+datatype+'_%Y%m%d_v??.cdf'
    elif instrument == 'grb':
        pathformat = 'gamma/grb_cdaweb/'+datatype+'/%Y/uy_'+datatype.split('_')[1]+'_'+datatype.split('_')[0]+'_%Y%m%d_v??.cdf'


    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange)

    out_files = []

    files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG['local_data_dir'], no_download=no_update)
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(out_files, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, notplot=notplot)
    
    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars



import cdflib

from pyspedas.analysis.time_clip import time_clip as tclip
from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pytplot import cdf_to_tplot

from .config import CONFIG

def load(trange=['2017-03-27', '2017-03-28'],
         pathformat=None,
         instrument='mgf',
         datatype='8sec',
         mode=None,
         site=None,
         model=None,
         level='l2',
         prefix='',
         suffix='',
         file_res=24*3600.,
         get_support_data=False,
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         uname=None,
         passwd=None,
         time_clip=False,
         version=None):
    """
    This function is not meant to be called directly; please see the instrument specific wrappers:
        pyspedas.erg.mgf()
        pyspedas.erg.hep()
        pyspedas.erg.orb()
        pyspedas.erg.lepe()
        pyspedas.erg.lepi()
        pyspedas.erg.mepe()
        pyspedas.erg.mepi()
        pyspedas.erg.pwe_ofa()
        pyspedas.erg.pwe_efd()
        pyspedas.erg.pwe_hfa()
        pyspedas.erg.xep()
    """

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat,
                              trange=trange, res=file_res)

    out_files = []

    files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG[
                     'local_data_dir'], no_download=no_update, last_version=True, username=uname, password=passwd)
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(out_files, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                         varformat=varformat, varnames=varnames, notplot=notplot)

    if notplot:
        if len(out_files) > 0:
            cdf_file = cdflib.CDF(out_files[-1])
            cdf_info = cdf_file.cdf_info()
            all_cdf_variables = cdf_info['rVariables'] + cdf_info['zVariables']
            gatt = cdf_file.globalattsget()
            for var in all_cdf_variables:
                t_plot_name = prefix + var + suffix
                if t_plot_name in tvars:
                    vatt = cdf_file.varattsget(var)
                    tvars[t_plot_name]['CDF'] = {'VATT':vatt,
                                                'GATT':gatt,
                                                'FILENAME':out_files}
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars



import os
import unittest
from pyspedas.utilities.data_exists import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_load_mag_data(self):
        mag_vars = pyspedas.solo.mag(time_clip=True)
        self.assertTrue(data_exists('B_RTN'))
        mag_vars = pyspedas.solo.mag(datatype='rtn-normal-1-minute')
        self.assertTrue(data_exists('B_RTN'))
        mag_vars = pyspedas.solo.mag(notplot=True, datatype='rtn-burst')
        self.assertTrue('B_RTN' in mag_vars)

    def test_load_mag_ll02_data(self):
        mag_vars = pyspedas.solo.mag(level='ll02', trange=['2020-08-04', '2020-08-05'])
        self.assertTrue(data_exists('B_RTN'))
        self.assertTrue(data_exists('B_SRF'))

    def test_load_epd_data(self):
        epd_vars = pyspedas.solo.epd()
        self.assertTrue(data_exists('Magnet_Rows_Flux'))
        self.assertTrue(data_exists('Integral_Rows_Flux'))
        self.assertTrue(data_exists('Magnet_Cols_Flux'))
        self.assertTrue(data_exists('Integral_Cols_Flux'))

    def test_load_rpw_data(self):
        rpw_vars = pyspedas.solo.rpw()
        self.assertTrue(data_exists('AVERAGE_NR'))
        self.assertTrue(data_exists('TEMPERATURE'))
        self.assertTrue(data_exists('FLUX_DENSITY1'))
        self.assertTrue(data_exists('FLUX_DENSITY2'))

    def test_load_swa_data(self):
        swa_vars = pyspedas.solo.swa()
        self.assertTrue(data_exists('eflux'))
        swa_vars = pyspedas.solo.swa(level='l2', datatype='eas1-nm3d-def')
        self.assertTrue(data_exists('SWA_EAS1_NM3D_DEF_Data'))
        swa_vars = pyspedas.solo.swa(notplot=True)
        self.assertTrue('eflux' in swa_vars)

    def test_load_swa_l1_data(self):
        swa_vars = pyspedas.solo.swa(level='l1', datatype='eas-padc')
        self.assertTrue(data_exists('SWA_EAS_BM_Data'))
        self.assertTrue(data_exists('SWA_EAS_MagDataUsed'))
        swa_vars = pyspedas.solo.swa(level='l1', datatype='his-pha', trange=['2020-06-03', '2020-06-04'])
        self.assertTrue(data_exists('HIS_PHA_EOQ_STEP'))

    def test_downloadonly(self):
        files = pyspedas.solo.mag(downloadonly=True)
        self.assertTrue(os.path.exists(files[0]))


if __name__ == '__main__':
    unittest.main()
